# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
"""Third-party libraries

This module contains packages that were developed outside Pyomo, but are
included in the Pyomo source tree to streamline deployment (e.g., they
are not easily installable via `pip` or `conda`, they are no longer
under active maintenance, or Pyomo requires custom patches).

"""



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

#
# Unit Tests for DerivativeVar Objects
#

import os
from os.path import abspath, dirname

import pyomo.common.unittest as unittest

from pyomo.environ import ConcreteModel, Var, Set, TransformationFactory
from pyomo.dae import ContinuousSet, DerivativeVar
from pyomo.dae.diffvar import DAE_Error

currdir = dirname(abspath(__file__)) + os.sep


class TestDerivativeVar(unittest.TestCase):
    # test valid declarations
    def test_valid(self):
        m = ConcreteModel()
        m.t = ContinuousSet(bounds=(0, 1))
        m.x = ContinuousSet(bounds=(5, 10))
        m.s = Set(dimen=1)
        m.v = Var(m.t)
        m.dv = DerivativeVar(m.v)
        m.dv2 = DerivativeVar(m.v, wrt=(m.t, m.t))

        self.assertTrue(isinstance(m.dv, Var))
        self.assertTrue(isinstance(m.dv, DerivativeVar))
        self.assertTrue(m.dv._wrt[0] is m.t)
        self.assertTrue(m.dv._sVar is m.v)
        self.assertTrue(m.v._derivative[('t',)]() is m.dv)
        self.assertTrue(m.dv.ctype is DerivativeVar)
        self.assertTrue(m.dv.index_set() is m.t)
        self.assertTrue(m.dv2._wrt[0] is m.t)
        self.assertTrue(m.dv2._wrt[1] is m.t)
        self.assertTrue(m.v._derivative[('t', 't')]() is m.dv2)
        self.assertTrue(m.dv.get_state_var() is m.v)
        self.assertTrue(m.dv2.get_state_var() is m.v)
        del m.dv
        del m.dv2
        del m.v

        m.v = Var(m.s, m.t)
        m.dv = DerivativeVar(m.v)
        m.dv2 = DerivativeVar(m.v, wrt=(m.t, m.t))
        self.assertTrue(isinstance(m.dv, Var))
        self.assertTrue(isinstance(m.dv, DerivativeVar))
        self.assertTrue(m.dv._wrt[0] is m.t)
        self.assertTrue(m.dv._sVar is m.v)
        self.assertTrue(m.v._derivative[('t',)]() is m.dv)
        self.assertTrue(m.dv.ctype is DerivativeVar)
        self.assertTrue(m.t in m.dv.index_set().set_tuple)
        self.assertTrue(m.s in m.dv.index_set().set_tuple)
        self.assertTrue(m.dv2._wrt[0] is m.t)
        self.assertTrue(m.dv2._wrt[1] is m.t)
        self.assertTrue(m.v._derivative[('t', 't')]() is m.dv2)
        del m.dv
        del m.dv2
        del m.v

        m.v = Var(m.x, m.t)
        m.dv = DerivativeVar(m.v, wrt=m.x)
        m.dv2 = DerivativeVar(m.v, wrt=m.t)
        m.dv3 = DerivativeVar(m.v, wrt=(m.t, m.x))
        m.dv4 = DerivativeVar(m.v, wrt=[m.t, m.t])
        self.assertTrue(isinstance(m.dv, Var))
        self.assertTrue(isinstance(m.dv, DerivativeVar))
        self.assertTrue(m.dv._wrt[0] is m.x)
        self.assertTrue(m.dv._sVar is m.v)
        self.assertTrue(m.v._derivative[('x',)]() is m.dv)
        self.assertTrue(m.v._derivative[('t',)]() is m.dv2)
        self.assertTrue(m.v._derivative[('t', 'x')]() is m.dv3)
        self.assertTrue(m.v._derivative[('t', 't')]() is m.dv4)
        self.assertTrue(m.dv.ctype is DerivativeVar)
        self.assertTrue(m.x in m.dv.index_set().set_tuple)
        self.assertTrue(m.t in m.dv.index_set().set_tuple)
        self.assertTrue(m.dv3._wrt[0] is m.t)
        self.assertTrue(m.dv3._wrt[1] is m.x)
        self.assertTrue(m.dv4._wrt[0] is m.t)
        self.assertTrue(m.dv4._wrt[1] is m.t)

    # test invalid declarations
    def test_invalid(self):
        m = ConcreteModel()
        m.t = ContinuousSet(bounds=(0, 1))
        m.x = ContinuousSet(bounds=(5, 10))
        m.s = Set(initialize=[1, 2, 3])
        m.v = Var(m.t)
        m.v2 = Var(m.s, m.t)
        m.v3 = Var(m.x, m.t)
        m.y = Var()

        # Not passing a Var as the first positional argument
        with self.assertRaises(DAE_Error):
            m.ds = DerivativeVar(m.s)

        # Specifying both option aliases
        with self.assertRaises(TypeError):
            m.dv = DerivativeVar(m.v, wrt=m.t, withrespectto=m.t)

        # Passing in Var not indexed by a ContinuousSet
        with self.assertRaises(DAE_Error):
            m.dy = DerivativeVar(m.y)

        # Not specifying 'wrt' when Var indexed by multiple ContinuousSets
        with self.assertRaises(DAE_Error):
            m.dv3 = DerivativeVar(m.v3)

        # 'wrt' is not a ContinuousSet
        with self.assertRaises(DAE_Error):
            m.dv2 = DerivativeVar(m.v2, wrt=m.s)

        with self.assertRaises(DAE_Error):
            m.dv2 = DerivativeVar(m.v2, wrt=(m.t, m.s))

        # Specified ContinuousSet does not index the Var
        with self.assertRaises(DAE_Error):
            m.dv = DerivativeVar(m.v, wrt=m.x)

        with self.assertRaises(DAE_Error):
            m.dv2 = DerivativeVar(m.v2, wrt=[m.t, m.x])

        # Declaring the same derivative twice
        m.dvdt = DerivativeVar(m.v)
        with self.assertRaises(DAE_Error):
            m.dvdt2 = DerivativeVar(m.v)

        m.dv2dt = DerivativeVar(m.v2, wrt=m.t)
        with self.assertRaises(DAE_Error):
            m.dv2dt2 = DerivativeVar(m.v2, wrt=m.t)

        m.dv3 = DerivativeVar(m.v3, wrt=(m.x, m.x))
        with self.assertRaises(DAE_Error):
            m.dv4 = DerivativeVar(m.v3, wrt=(m.x, m.x))

    # test DerivativeVar reclassification after discretization
    def test_reclassification(self):
        m = ConcreteModel()
        m.t = ContinuousSet(bounds=(0, 1))
        m.x = ContinuousSet(bounds=(5, 10))
        m.s = Set(initialize=[1, 2, 3])
        m.v = Var(m.t)
        m.v2 = Var(m.s, m.t)
        m.v3 = Var(m.x, m.t)

        m.dv = DerivativeVar(m.v)
        m.dv2 = DerivativeVar(m.v2, wrt=(m.t, m.t))
        m.dv3 = DerivativeVar(m.v3, wrt=m.x)

        TransformationFactory('dae.finite_difference').apply_to(m, wrt=m.t)

        self.assertTrue(m.dv.ctype is Var)
        self.assertTrue(m.dv2.ctype is Var)
        self.assertTrue(m.dv.is_fully_discretized())
        self.assertTrue(m.dv2.is_fully_discretized())
        self.assertTrue(m.dv3.ctype is DerivativeVar)
        self.assertFalse(m.dv3.is_fully_discretized())

        TransformationFactory('dae.collocation').apply_to(m, wrt=m.x)
        self.assertTrue(m.dv3.ctype is Var)
        self.assertTrue(m.dv3.is_fully_discretized())


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for Elements of a Model
#
# TestSimpleVar                Class for testing single variables
# TestArrayVar                Class for testing array of variables
#

import os
from os.path import abspath, dirname

currdir = dirname(abspath(__file__)) + os.sep

from io import StringIO

import pyomo.common.unittest as unittest
from pyomo.common.log import LoggingIntercept

from pyomo.core.base import IntegerSet
from pyomo.core.expr.numeric_expr import (
    NPV_ProductExpression,
    NPV_MaxExpression,
    NPV_MinExpression,
)
from pyomo.core.staleflag import StaleFlagManager
from pyomo.environ import (
    AbstractModel,
    ConcreteModel,
    Set,
    Param,
    Var,
    VarList,
    RangeSet,
    Suffix,
    Expression,
    NonPositiveReals,
    PositiveReals,
    Reals,
    RealSet,
    NonNegativeReals,
    Integers,
    Binary,
    value,
)
from pyomo.core.base.units_container import units, pint_available, UnitsError


class TestVarData(unittest.TestCase):
    def test_lower_bound(self):
        m = ConcreteModel()
        m.x = Var()
        m.p = Param(mutable=True, initialize=2)
        self.assertIsNone(m.x.lower)
        m.x.domain = NonNegativeReals
        self.assertIs(type(m.x.lower), int)
        self.assertEqual(value(m.x.lower), 0)
        m.x.domain = Reals
        m.x.setlb(5 * m.p)
        self.assertIs(type(m.x.lower), NPV_ProductExpression)
        self.assertEqual(value(m.x.lower), 10)
        m.x.domain = NonNegativeReals
        self.assertIs(type(m.x.lower), NPV_MaxExpression)
        self.assertEqual(value(m.x.lower), 10)
        with self.assertRaisesRegex(
            ValueError,
            "Potentially variable input of type 'ScalarVar' "
            "supplied as lower bound for variable 'x'",
        ):
            m.x.setlb(m.x)

    def test_lower_bound_setter(self):
        m = ConcreteModel()
        m.x = Var()
        self.assertIsNone(m.x.lb)
        m.x.lb = 1
        self.assertEqual(m.x.lb, 1)
        m.x.lower = 2
        self.assertEqual(m.x.lb, 2)
        m.x.setlb(3)
        self.assertEqual(m.x.lb, 3)

        m.y = Var([1])
        self.assertIsNone(m.y[1].lb)
        m.y[1].lb = 1
        self.assertEqual(m.y[1].lb, 1)
        m.y[1].lower = 2
        self.assertEqual(m.y[1].lb, 2)
        m.y[1].setlb(3)
        self.assertEqual(m.y[1].lb, 3)

    def test_upper_bound(self):
        m = ConcreteModel()
        m.x = Var()
        m.p = Param(mutable=True, initialize=2)
        self.assertIsNone(m.x.upper)
        m.x.domain = NonPositiveReals
        self.assertIs(type(m.x.upper), int)
        self.assertEqual(value(m.x.upper), 0)
        m.x.domain = Reals
        m.x.setub(-5 * m.p)
        self.assertIs(type(m.x.upper), NPV_ProductExpression)
        self.assertEqual(value(m.x.upper), -10)
        m.x.domain = NonPositiveReals
        self.assertIs(type(m.x.upper), NPV_MinExpression)
        self.assertEqual(value(m.x.upper), -10)
        with self.assertRaisesRegex(
            ValueError,
            "Potentially variable input of type 'ScalarVar' "
            "supplied as upper bound for variable 'x'",
        ):
            m.x.setub(m.x)

    def test_upper_bound_setter(self):
        m = ConcreteModel()
        m.x = Var()
        self.assertIsNone(m.x.ub)
        m.x.ub = 1
        self.assertEqual(m.x.ub, 1)
        m.x.upper = 2
        self.assertEqual(m.x.ub, 2)
        m.x.setub(3)
        self.assertEqual(m.x.ub, 3)

        m.y = Var([1])
        self.assertIsNone(m.y[1].ub)
        m.y[1].ub = 1
        self.assertEqual(m.y[1].ub, 1)
        m.y[1].upper = 2
        self.assertEqual(m.y[1].ub, 2)
        m.y[1].setub(3)
        self.assertEqual(m.y[1].ub, 3)

    def test_lb(self):
        m = ConcreteModel()
        m.x = Var()
        self.assertEqual(m.x.lb, None)
        m.x.domain = NonNegativeReals
        self.assertEqual(m.x.lb, 0)
        m.x.lb = float('inf')
        with self.assertRaisesRegex(
            ValueError, r'invalid non-finite lower bound \(inf\)'
        ):
            m.x.lb
        m.x.lb = float('nan')
        with self.assertRaisesRegex(
            ValueError, r'invalid non-finite lower bound \(nan\)'
        ):
            m.x.lb

    def test_ub(self):
        m = ConcreteModel()
        m.x = Var()
        self.assertEqual(m.x.ub, None)
        m.x.domain = NonPositiveReals
        self.assertEqual(m.x.ub, 0)
        m.x.ub = float('-inf')
        with self.assertRaisesRegex(
            ValueError, r'invalid non-finite upper bound \(-inf\)'
        ):
            m.x.ub
        m.x.ub = float('nan')
        with self.assertRaisesRegex(
            ValueError, r'invalid non-finite upper bound \(nan\)'
        ):
            m.x.ub

    def test_bounds(self):
        m = ConcreteModel()
        m.x = Var()
        lb, ub = m.x.bounds
        self.assertEqual(lb, None)
        self.assertEqual(ub, None)
        m.x.domain = NonNegativeReals
        lb, ub = m.x.bounds
        self.assertEqual(lb, 0)
        self.assertEqual(ub, None)
        m.x.lb = float('inf')
        with self.assertRaisesRegex(
            ValueError, r'invalid non-finite lower bound \(inf\)'
        ):
            lb, ub = m.x.bounds
        m.x.lb = float('nan')
        with self.assertRaisesRegex(
            ValueError, r'invalid non-finite lower bound \(nan\)'
        ):
            lb, ub = m.x.bounds

        m.x.lb = None
        m.x.domain = NonPositiveReals
        lb, ub = m.x.bounds
        self.assertEqual(lb, None)
        self.assertEqual(ub, 0)
        m.x.ub = float('-inf')
        with self.assertRaisesRegex(
            ValueError, r'invalid non-finite upper bound \(-inf\)'
        ):
            lb, ub = m.x.bounds
        m.x.ub = float('nan')
        with self.assertRaisesRegex(
            ValueError, r'invalid non-finite upper bound \(nan\)'
        ):
            lb, ub = m.x.bounds


class PyomoModel(unittest.TestCase):
    def setUp(self):
        self.model = AbstractModel()
        self.instance = None

    def tearDown(self):
        self.model = None
        self.instance = None

    def construct(self, filename=None):
        if filename is not None:
            self.instance = self.model.create_instance(filename)
        else:
            self.instance = self.model.create_instance()


class TestSimpleVar(PyomoModel):
    def setUp(self):
        #
        # Create Model
        #
        PyomoModel.setUp(self)

    def test_fixed_attr(self):
        """Test fixed attribute"""
        self.model.x = Var()
        self.instance = self.model.create_instance()
        self.instance.x.fixed = True
        self.assertEqual(self.instance.x.fixed, True)

    def Xtest_setlb_nondata_expression(self):
        model = ConcreteModel()
        model.x = Var()
        model.e = Expression()
        with self.assertRaises(ValueError):
            model.x.setlb(model.e)
        model.e.expr = 1.0
        with self.assertRaises(ValueError):
            model.x.setlb(model.e)
        model.y = Var()
        with self.assertRaises(ValueError):
            model.x.setlb(model.y)
        model.y.value = 1.0
        with self.assertRaises(ValueError):
            model.x.setlb(model.y)
        model.y.fix()
        with self.assertRaises(ValueError):
            model.x.setlb(model.y + 1)

    def Xtest_setub_nondata_expression(self):
        model = ConcreteModel()
        model.x = Var()
        model.e = Expression()
        with self.assertRaises(ValueError):
            model.x.setub(model.e)
        model.e.expr = 1.0
        with self.assertRaises(ValueError):
            model.x.setub(model.e)
        model.y = Var()
        with self.assertRaises(ValueError):
            model.x.setub(model.y)
        model.y.value = 1.0
        with self.assertRaises(ValueError):
            model.x.setub(model.y)
        model.y.fix()
        with self.assertRaises(ValueError):
            model.x.setub(model.y + 1)

    def Xtest_setlb_data_expression(self):
        model = ConcreteModel()
        model.x = Var()
        model.p = Param(mutable=True)
        model.x.setlb(model.p)
        model.x.setlb(model.p**2 + 1)
        model.p.value = 1.0
        model.x.setlb(model.p)
        model.x.setlb(model.p**2)
        model.x.setlb(1.0)

    def Xtest_setub_data_expression(self):
        model = ConcreteModel()
        model.x = Var()
        model.p = Param(mutable=True)
        model.x.setub(model.p)
        model.x.setub(model.p**2 + 1)
        model.p.value = 1.0
        model.x.setub(model.p)
        model.x.setub(model.p**2)
        model.x.setub(1.0)

    def test_setlb_indexed(self):
        """Test setlb variables method"""
        self.model.B = RangeSet(4)
        self.model.y = Var(self.model.B, dense=True)

        self.instance = self.model.create_instance()
        self.assertEqual(len(self.instance.y) > 0, True)
        for a in self.instance.y:
            self.assertEqual(self.instance.y[a].lb, None)
        self.instance.y.setlb(1)
        for a in self.instance.y:
            self.assertEqual(self.instance.y[a].lb, 1)
        self.instance.y.setlb(None)
        for a in self.instance.y:
            self.assertEqual(self.instance.y[a].lb, None)

    def test_setub_indexed(self):
        """Test setub variables method"""
        self.model.B = RangeSet(4)
        self.model.y = Var(self.model.B, dense=True)

        self.instance = self.model.create_instance()
        self.assertEqual(len(self.instance.y) > 0, True)
        for a in self.instance.y:
            self.assertEqual(self.instance.y[a].ub, None)
        self.instance.y.setub(1)
        for a in self.instance.y:
            self.assertEqual(self.instance.y[a].ub, 1)
        self.instance.y.setub(None)
        for a in self.instance.y:
            self.assertEqual(self.instance.y[a].ub, None)

    def test_fix_all(self):
        """Test fix all variables method"""
        self.model.B = RangeSet(4)
        self.model.x = Var()
        self.model.y = Var(self.model.B, dense=True)

        self.instance = self.model.create_instance()
        self.instance.fix_all_vars()
        self.assertEqual(self.instance.x.fixed, True)

        for a in self.instance.y:
            self.assertEqual(self.instance.y[a].fixed, True)

    def test_unfix_all(self):
        """Test unfix all variables method"""
        self.model.B = RangeSet(4)
        self.model.x = Var()
        self.model.y = Var(self.model.B)

        self.instance = self.model.create_instance()
        self.instance.x.fixed = True
        for a in self.instance.B:
            self.instance.y[a].fixed = True
        self.instance.unfix_all_vars()

        self.assertEqual(self.instance.x.fixed, False)
        for a in self.instance.B:
            self.assertEqual(self.instance.y[a].fixed, False)

    def test_fix_indexed(self):
        """Test fix variables method"""
        self.model.B = RangeSet(4)
        self.model.y = Var(self.model.B, dense=True)

        self.instance = self.model.create_instance()
        self.assertEqual(len(self.instance.y) > 0, True)
        for a in self.instance.y:
            self.assertEqual(self.instance.y[a].value, None)
            self.assertEqual(self.instance.y[a].fixed, False)
        self.instance.y.fix()
        for a in self.instance.y:
            self.assertEqual(self.instance.y[a].value, None)
            self.assertEqual(self.instance.y[a].fixed, True)
        self.instance.y.free()
        for a in self.instance.y:
            self.assertEqual(self.instance.y[a].value, None)
            self.assertEqual(self.instance.y[a].fixed, False)
        self.instance.y.fix(1)
        for a in self.instance.y:
            self.assertEqual(self.instance.y[a].value, 1)
            self.assertEqual(self.instance.y[a].fixed, True)
        self.instance.y.unfix()
        for a in self.instance.y:
            self.assertEqual(self.instance.y[a].value, 1)
            self.assertEqual(self.instance.y[a].fixed, False)
        self.instance.y.fix(None)
        for a in self.instance.y:
            self.assertEqual(self.instance.y[a].value, None)
            self.assertEqual(self.instance.y[a].fixed, True)
        self.instance.y.unfix()
        for a in self.instance.y:
            self.assertEqual(self.instance.y[a].value, None)
            self.assertEqual(self.instance.y[a].fixed, False)

        self.instance.y[1].fix()
        self.assertEqual(self.instance.y[1].value, None)
        self.assertEqual(self.instance.y[1].fixed, True)
        self.instance.y[1].free()
        self.assertEqual(self.instance.y[1].value, None)
        self.assertEqual(self.instance.y[1].fixed, False)
        self.instance.y[1].fix(value=1)
        self.assertEqual(self.instance.y[1].value, 1)
        self.assertEqual(self.instance.y[1].fixed, True)
        self.instance.y[1].unfix()
        self.assertEqual(self.instance.y[1].value, 1)
        self.assertEqual(self.instance.y[1].fixed, False)
        self.instance.y[1].fix(value=None)
        self.assertEqual(self.instance.y[1].value, None)
        self.assertEqual(self.instance.y[1].fixed, True)

    def test_unfix_indexed(self):
        """Test unfix variables method"""
        self.model.B = RangeSet(4)
        self.model.y = Var(self.model.B)

        self.instance = self.model.create_instance()
        for a in self.instance.B:
            self.instance.y[a].fixed = True
        self.instance.unfix_all_vars()

        for a in self.instance.B:
            self.assertEqual(self.instance.y[a].fixed, False)

    def test_fix_nonindexed(self):
        """Test fix variables method"""
        self.model.B = RangeSet(4)
        self.model.x = Var()

        self.instance = self.model.create_instance()
        self.assertEqual(self.instance.x.value, None)
        self.assertEqual(self.instance.x.fixed, False)
        self.instance.x.fix()
        self.assertEqual(self.instance.x.value, None)
        self.assertEqual(self.instance.x.fixed, True)
        self.instance.x.free()
        self.assertEqual(self.instance.x.value, None)
        self.assertEqual(self.instance.x.fixed, False)
        self.instance.x.fix(1)
        self.assertEqual(self.instance.x.value, 1)
        self.assertEqual(self.instance.x.fixed, True)
        self.instance.x.unfix()
        self.assertEqual(self.instance.x.value, 1)
        self.assertEqual(self.instance.x.fixed, False)
        self.instance.x.fix(None)
        self.assertEqual(self.instance.x.value, None)
        self.assertEqual(self.instance.x.fixed, True)

    def test_unfix_nonindexed(self):
        """Test unfix variables method"""
        self.model.B = RangeSet(4)
        self.model.x = Var()
        self.model.y = Var(self.model.B)

        self.instance = self.model.create_instance()
        self.instance.x.fixed = True
        self.instance.x.unfix()

        self.assertEqual(self.instance.x.fixed, False)

    def test_value_attr(self):
        """Test value attribute"""
        self.model.x = Var()
        self.instance = self.model.create_instance()
        self.instance.x.value = 3.5
        self.assertEqual(self.instance.x.value, 3.5)

    def test_domain_attr(self):
        """Test domain attribute"""
        self.model.x = Var()
        self.instance = self.model.create_instance()
        self.instance.x.domain = Integers
        self.assertEqual(type(self.instance.x.domain), IntegerSet)
        self.assertEqual(self.instance.x.is_integer(), True)
        self.assertEqual(self.instance.x.is_binary(), False)
        self.assertEqual(self.instance.x.is_continuous(), False)

    def test_name_attr(self):
        """Test name attribute"""
        #
        # A user would never need to do this, but this
        # attribute is needed within Pyomo
        #
        self.model.x = Var()
        self.model.x._name = "foo"
        self.assertEqual(self.model.x.name, "foo")

    def test_lb_attr1(self):
        """Test lb attribute"""
        self.model.x = Var()
        self.instance = self.model.create_instance()
        self.instance.x.setlb(-1.0)
        self.assertEqual(value(self.instance.x.lb), -1.0)

    def test_lb_attr2(self):
        """Test lb attribute"""
        self.model.x = Var(within=NonNegativeReals, bounds=(-1, 2))
        self.instance = self.model.create_instance()
        self.assertEqual(value(self.instance.x.lb), 0.0)
        self.assertEqual(value(self.instance.x.ub), 2.0)

    def test_lb_attr3(self):
        """Test lb attribute"""
        self.model.p = Param(mutable=True, initialize=1)
        self.model.x = Var(within=NonNegativeReals, bounds=(self.model.p, None))
        self.instance = self.model.create_instance()
        self.assertEqual(value(self.instance.x.lb), 1.0)
        self.instance.p = 2
        self.assertEqual(value(self.instance.x.lb), 2.0)

    def test_ub_attr1(self):
        """Test ub attribute"""
        self.model.x = Var()
        self.instance = self.model.create_instance()
        self.instance.x.setub(1.0)
        self.assertEqual(value(self.instance.x.ub), 1.0)

    def test_ub_attr2(self):
        """Test ub attribute"""
        self.model.x = Var(within=NonPositiveReals, bounds=(-2, 1))
        self.instance = self.model.create_instance()
        self.assertEqual(value(self.instance.x.lb), -2.0)
        self.assertEqual(value(self.instance.x.ub), 0.0)

    def test_within_option(self):
        """Test within option"""
        self.model.x = Var(within=Reals)
        self.construct()
        self.assertEqual(type(self.instance.x.domain), RealSet)
        self.assertEqual(self.instance.x.is_integer(), False)
        self.assertEqual(self.instance.x.is_binary(), False)
        self.assertEqual(self.instance.x.is_continuous(), True)

    def test_bounds_option1(self):
        """Test bounds option"""

        def x_bounds(model):
            return (-1.0, 1.0)

        self.model.x = Var(bounds=x_bounds)
        self.instance = self.model.create_instance()
        self.assertEqual(value(self.instance.x.lb), -1.0)
        self.assertEqual(value(self.instance.x.ub), 1.0)

    def test_bounds_option2(self):
        """Test bounds option"""
        self.model.x = Var(bounds=(-1.0, 1.0))
        self.instance = self.model.create_instance()
        self.assertEqual(value(self.instance.x.lb), -1.0)
        self.assertEqual(value(self.instance.x.ub), 1.0)

    def test_rule_option(self):
        """Test rule option"""

        def x_init(model):
            return 1.3

        self.model.x = Var(initialize=x_init)
        self.instance = self.model.create_instance()
        self.assertEqual(self.instance.x.value, 1.3)

    def test_initialize_with_function(self):
        """Test initialize option with an initialization rule"""

        def init_rule(model):
            return 1.3

        self.model.x = Var(initialize=init_rule)
        self.instance = self.model.create_instance()
        self.assertEqual(self.instance.x.value, 1.3)
        self.instance.x = 1
        self.assertEqual(self.instance.x.value, 1)

    def test_initialize_with_dict(self):
        """Test initialize option with a dictionary"""
        self.model.x = Var(initialize={None: 1.3})
        self.instance = self.model.create_instance()
        self.assertEqual(self.instance.x.value, 1.3)
        self.instance.x = 1
        self.assertEqual(self.instance.x.value, 1)

    def test_initialize_with_const(self):
        """Test initialize option with a constant"""
        self.model.x = Var(initialize=1.3)
        self.instance = self.model.create_instance()
        self.assertEqual(self.instance.x.value, 1.3)
        self.instance.x = 1
        self.assertEqual(self.instance.x.value, 1)

    def test_without_initial_value(self):
        """Test default initial value"""
        self.model.x = Var()
        self.instance = self.model.create_instance()
        self.assertEqual(self.instance.x.value, None)
        self.instance.x = 6
        self.assertEqual(self.instance.x.value, 6)

    def test_dim(self):
        """Test dim method"""
        self.model.x = Var()
        self.instance = self.model.create_instance()
        self.assertEqual(self.instance.x.dim(), 0)

    def test_keys(self):
        """Test keys method"""
        self.model.x = Var()
        self.instance = self.model.create_instance()
        self.assertEqual(list(self.instance.x.keys()), [None])
        self.assertEqual(id(self.instance.x), id(self.instance.x[None]))

    def test_len(self):
        """Test len method"""
        self.model.x = Var()
        self.assertEqual(len(self.model.x), 0)
        self.instance = self.model.create_instance()
        self.assertEqual(len(self.instance.x), 1)

    def test_value(self):
        """Check the value of the variable"""
        self.model.x = Var(initialize=3.3)
        self.instance = self.model.create_instance()
        tmp = value(self.instance.x.value)
        self.assertEqual(type(tmp), float)
        self.assertEqual(tmp, 3.3)
        tmp = float(self.instance.x.value)
        self.assertEqual(type(tmp), float)
        self.assertEqual(tmp, 3.3)
        tmp = int(self.instance.x.value)
        self.assertEqual(type(tmp), int)
        self.assertEqual(tmp, 3)


class TestArrayVar(TestSimpleVar):
    def setUp(self):
        #
        # Create Model
        #
        PyomoModel.setUp(self)
        self.model.A = Set(initialize=[1, 2])

    def test_fixed_attr(self):
        """Test fixed attribute"""
        self.model.x = Var(self.model.A)
        self.model.y = Var(self.model.A)
        self.instance = self.model.create_instance()
        self.instance.x.fixed = True
        self.assertEqual(self.instance.x[1].fixed, False)
        self.instance.y[1].fixed = True
        self.assertEqual(self.instance.y[1].fixed, True)

    def test_value_attr(self):
        """Test value attribute"""
        self.model.x = Var(self.model.A, dense=True)
        self.model.y = Var(self.model.A, dense=True)
        self.instance = self.model.create_instance()
        try:
            self.instance.x = 3.5
            self.fail("Expected ValueError")
        except ValueError:
            pass
        self.instance.y[1] = 3.5
        self.assertEqual(self.instance.y[1].value, 3.5)

    # def test_lb_attr(self):
    # """Test lb attribute"""
    # self.model.x = Var(self.model.A)
    # self.instance = self.model.create_instance()
    # self.instance.x.setlb(-1.0)
    # self.assertEqual(value(self.instance.x[1].lb), -1.0)

    # def test_ub_attr(self):
    # """Test ub attribute"""
    # self.model.x = Var(self.model.A)
    # self.instance = self.model.create_instance()
    # self.instance.x.setub(1.0)
    # self.assertEqual(value(self.instance.x[1].ub), 1.0)

    def test_initialize_with_function(self):
        """Test initialize option with an initialization rule"""

        def init_rule(model, key):
            i = key + 11
            return key == 1 and 1.3 or 2.3

        self.model.x = Var(self.model.A, initialize=init_rule)
        self.instance = self.model.create_instance()
        self.assertEqual(self.instance.x[1].value, 1.3)
        self.assertEqual(self.instance.x[2].value, 2.3)
        self.instance.x[1] = 1
        self.instance.x[2] = 2
        self.assertEqual(self.instance.x[1].value, 1)
        self.assertEqual(self.instance.x[2].value, 2)

    def test_initialize_with_dict(self):
        """Test initialize option with a dictionary"""
        self.model.x = Var(self.model.A, initialize={1: 1.3, 2: 2.3})
        self.instance = self.model.create_instance()
        self.assertEqual(self.instance.x[1].value, 1.3)
        self.assertEqual(self.instance.x[2].value, 2.3)
        self.instance.x[1] = 1
        self.instance.x[2] = 2
        self.assertEqual(self.instance.x[1].value, 1)
        self.assertEqual(self.instance.x[2].value, 2)

    def test_initialize_with_subdict(self):
        """Test initialize option method with a dictionary of subkeys"""
        self.model.x = Var(self.model.A, initialize={1: 1.3})
        self.instance = self.model.create_instance()
        self.assertEqual(self.instance.x[1].value, 1.3)
        self.assertEqual(self.instance.x[2].value, None)
        self.instance.x[1] = 1
        self.instance.x[2] = 2
        self.assertEqual(self.instance.x[1].value, 1)
        self.assertEqual(self.instance.x[2].value, 2)

    def test_initialize_with_const(self):
        """Test initialize option with a constant"""
        self.model.x = Var(self.model.A, initialize=3)
        self.instance = self.model.create_instance()
        self.assertEqual(self.instance.x[1].value, 3)
        self.assertEqual(self.instance.x[2].value, 3)
        self.instance.x[1] = 1
        self.instance.x[2] = 2
        self.assertEqual(self.instance.x[1].value, 1)
        self.assertEqual(self.instance.x[2].value, 2)

    def test_without_initial_value(self):
        """Test default initial value"""
        self.model.x = Var(self.model.A)
        self.instance = self.model.create_instance()
        self.assertEqual(self.instance.x[1].value, None)
        self.assertEqual(self.instance.x[2].value, None)
        self.instance.x[1] = 5
        self.instance.x[2] = 6
        self.assertEqual(self.instance.x[1].value, 5)
        self.assertEqual(self.instance.x[2].value, 6)

    def test_bounds_option1(self):
        """Test bounds option"""

        def x_bounds(model, i):
            return (-1.0, 1.0)

        self.model.x = Var(self.model.A, bounds=x_bounds)
        self.instance = self.model.create_instance()
        self.assertEqual(value(self.instance.x[1].lb), -1.0)
        self.assertEqual(value(self.instance.x[1].ub), 1.0)

    def test_bounds_option2(self):
        """Test bounds option"""
        self.model.x = Var(self.model.A, bounds=(-1.0, 1.0))
        self.instance = self.model.create_instance()
        self.assertEqual(value(self.instance.x[1].lb), -1.0)
        self.assertEqual(value(self.instance.x[1].ub), 1.0)

    def test_rule_option(self):
        """Test rule option"""

        def x_init(model, i):
            return 1.3

        self.model.x = Var(self.model.A, initialize=x_init)
        self.instance = self.model.create_instance()
        self.assertEqual(self.instance.x[1].value, 1.3)

    def test_dim(self):
        """Test dim method"""
        self.model.x = Var(self.model.A)
        self.instance = self.model.create_instance()
        self.assertEqual(self.instance.x.dim(), 1)

    def test_keys(self):
        """Test keys method"""
        self.model.x = Var(self.model.A, dense=False)
        self.model.y = Var(self.model.A, dense=True)
        self.model.z = Var(self.model.A)
        self.instance = self.model.create_instance()
        self.assertEqual(set(self.instance.x.keys()), set())
        self.assertEqual(set(self.instance.y.keys()), set([1, 2]))
        self.assertEqual(set(self.instance.z.keys()), set([1, 2]))

    def test_len(self):
        """Test len method"""
        self.model.x = Var(self.model.A, dense=False)
        self.model.y = Var(self.model.A, dense=True)
        self.model.z = Var(self.model.A)
        self.instance = self.model.create_instance()
        self.assertEqual(len(self.instance.x), 0)
        self.assertEqual(len(self.instance.y), 2)
        self.assertEqual(len(self.instance.z), 2)

    def test_value(self):
        """Check the value of the variable"""
        self.model.x = Var(self.model.A, initialize=3.3)
        self.instance = self.model.create_instance()
        tmp = value(self.instance.x[1].value)
        self.assertEqual(type(tmp), float)
        self.assertEqual(tmp, 3.3)
        tmp = float(self.instance.x[1].value)
        self.assertEqual(type(tmp), float)
        self.assertEqual(tmp, 3.3)
        tmp = int(self.instance.x[1].value)
        self.assertEqual(type(tmp), int)
        self.assertEqual(tmp, 3)

    def test_var_domain_setter(self):
        m = ConcreteModel()
        m.x = Var([1, 2, 3])
        self.assertIs(m.x[1].domain, Reals)
        self.assertIs(m.x[2].domain, Reals)
        self.assertIs(m.x[3].domain, Reals)
        m.x.domain = Integers
        self.assertIs(m.x[1].domain, Integers)
        self.assertIs(m.x[2].domain, Integers)
        self.assertIs(m.x[3].domain, Integers)
        m.x.domain = lambda m, i: PositiveReals
        self.assertIs(m.x[1].domain, PositiveReals)
        self.assertIs(m.x[2].domain, PositiveReals)
        self.assertIs(m.x[3].domain, PositiveReals)
        m.x.domain = {1: Reals, 2: NonPositiveReals, 3: NonNegativeReals}
        self.assertIs(m.x[1].domain, Reals)
        self.assertIs(m.x[2].domain, NonPositiveReals)
        self.assertIs(m.x[3].domain, NonNegativeReals)
        m.x.domain = {2: Integers}
        self.assertIs(m.x[1].domain, Reals)
        self.assertIs(m.x[2].domain, Integers)
        self.assertIs(m.x[3].domain, NonNegativeReals)
        with (
            LoggingIntercept() as LOG,
            self.assertRaisesRegex(
                TypeError,
                'Cannot create a Set from data that does not support __contains__.  '
                'Expected set-like object supporting collections.abc.Collection '
                "interface, but received 'NoneType'",
            ),
        ):
            m.x.domain = {1: None, 2: None, 3: None}
        self.assertIn(
            '{1: None, 2: None, 3: None} is not a valid domain. '
            'Variable domains must be an instance of a Pyomo Set or '
            'convertible to a Pyomo Set.',
            LOG.getvalue(),
        )


class TestVarList(PyomoModel):
    def setUp(self):
        #
        # Create Model
        #
        PyomoModel.setUp(self)

    def test_fixed_attr(self):
        """Test fixed attribute"""
        self.model.x = VarList()
        self.model.y = VarList()
        self.instance = self.model.create_instance()
        self.instance.x.add()
        self.instance.x.add()
        self.instance.x.add()
        self.instance.y.add()
        self.instance.y.add()
        self.instance.y.add()
        self.instance.x.fixed = True
        self.assertEqual(self.instance.x[1].fixed, False)
        self.instance.y[1].fixed = True
        self.assertEqual(self.instance.y[1].fixed, True)

    def test_value_attr(self):
        """Test value attribute"""
        self.model.x = VarList()
        self.model.y = VarList()
        self.instance = self.model.create_instance()
        self.instance.x.add()
        self.instance.x.add()
        self.instance.x.add()
        self.instance.y.add()
        self.instance.y.add()
        self.instance.y.add()
        try:
            self.instance.x = 3.5
            self.fail("Expected ValueError")
        except ValueError:
            pass
        self.instance.y[1] = 3.5
        self.assertEqual(self.instance.y[1].value, 3.5)

    def test_initialize_with_function(self):
        """Test initialize option with an initialization rule"""

        def init_rule(model, key):
            i = key + 11
            return key == 1 and 1.3 or 2.3

        self.model.x = VarList(initialize=init_rule)
        self.instance = self.model.create_instance()
        self.instance.x.add()
        self.instance.x.add()
        self.instance.x.add()
        self.assertEqual(self.instance.x[1].value, 1.3)
        self.assertEqual(self.instance.x[2].value, 2.3)
        self.instance.x[1] = 1
        self.instance.x[2] = 2
        self.assertEqual(self.instance.x[1].value, 1)
        self.assertEqual(self.instance.x[2].value, 2)

    def test_initialize_with_dict(self):
        """Test initialize option with a dictionary"""
        self.model.x = VarList(initialize={1: 1.3, 2: 2.3})
        self.instance = self.model.create_instance()
        self.assertEqual(self.instance.x[1].value, 1.3)
        self.assertEqual(self.instance.x[2].value, 2.3)
        self.instance.x[1] = 1
        self.instance.x[2] = 2
        self.assertEqual(self.instance.x[1].value, 1)
        self.assertEqual(self.instance.x[2].value, 2)

    def test_initialize_with_bad_dict(self):
        """Test initialize option with a dictionary of subkeys"""
        self.model.x = VarList(initialize={0: 1.3})
        self.assertRaisesRegex(
            KeyError,
            ".*Index '0' is not valid for indexed component 'x'",
            self.model.create_instance,
        )

    def test_initialize_with_const(self):
        """Test initialize option with a constant"""
        self.model.x = VarList(initialize=3)
        self.instance = self.model.create_instance()
        self.instance.x.add()
        self.instance.x.add()
        self.instance.x.add()
        self.assertEqual(self.instance.x[1].value, 3)
        self.assertEqual(self.instance.x[2].value, 3)
        self.instance.x[1] = 1
        self.instance.x[2] = 2
        self.assertEqual(self.instance.x[1].value, 1)
        self.assertEqual(self.instance.x[2].value, 2)

    def test_without_initial_value(self):
        """Test default initialization method"""
        self.model.x = VarList()
        self.instance = self.model.create_instance()
        self.instance.x.add()
        self.instance.x.add()
        self.instance.x.add()
        self.assertEqual(self.instance.x[1].value, None)
        self.assertEqual(self.instance.x[2].value, None)
        self.instance.x[1] = 5
        self.instance.x[2] = 6
        self.assertEqual(self.instance.x[1].value, 5)
        self.assertEqual(self.instance.x[2].value, 6)

    def test_bounds_option1(self):
        """Test bounds option"""

        def x_bounds(model, i):
            return (-1.0, 1.0)

        self.model.x = VarList(bounds=x_bounds)
        self.instance = self.model.create_instance()
        self.instance.x.add()
        self.instance.x.add()
        self.assertEqual(value(self.instance.x[1].lb), -1.0)
        self.assertEqual(value(self.instance.x[1].ub), 1.0)

    def test_bounds_option2(self):
        """Test bounds option"""
        self.model.x = VarList(bounds=(-1.0, 1.0))
        self.instance = self.model.create_instance()
        self.instance.x.add()
        self.instance.x.add()
        self.assertEqual(value(self.instance.x[1].lb), -1.0)
        self.assertEqual(value(self.instance.x[1].ub), 1.0)

    def test_rule_option(self):
        """Test rule option"""

        def x_init(model, i):
            return 1.3

        self.model.x = VarList(initialize=x_init)
        self.instance = self.model.create_instance()
        self.instance.x.add()
        self.instance.x.add()
        self.assertEqual(self.instance.x[1].value, 1.3)

    def test_domain1(self):
        self.model.x = VarList(domain=NonNegativeReals)
        self.instance = self.model.create_instance()
        self.instance.x.add()
        self.instance.x.add()
        self.assertEqual(str(self.instance.x[1].domain), str(NonNegativeReals))
        self.assertEqual(str(self.instance.x[2].domain), str(NonNegativeReals))
        self.instance.x[1].domain = Integers
        self.assertEqual(str(self.instance.x[1].domain), str(Integers))

    def test_domain2(self):
        def x_domain(model, i):
            if i == 1:
                return NonNegativeReals
            elif i == 2:
                return Reals
            elif i == 3:
                return Integers

        self.model.x = VarList(domain=x_domain)
        self.instance = self.model.create_instance()
        self.instance.x.add()
        self.instance.x.add()
        self.instance.x.add()
        self.assertEqual(str(self.instance.x[1].domain), str(NonNegativeReals))
        self.assertEqual(str(self.instance.x[2].domain), str(Reals))
        self.assertEqual(str(self.instance.x[3].domain), str(Integers))
        try:
            self.instance.x.domain
        except AttributeError:
            pass
        # test the property setter
        self.instance.x.domain = Binary
        self.assertEqual(str(self.instance.x[1].domain), str(Binary))
        self.assertEqual(str(self.instance.x[2].domain), str(Binary))
        self.assertEqual(str(self.instance.x[3].domain), str(Binary))

    # VarList doesn't handle generators yet
    @unittest.expectedFailure
    def test_domain3(self):
        def x_domain(model):
            yield NonNegativeReals
            yield Reals
            yield Integers

        self.model.x = VarList(domain=x_domain)
        self.instance = self.model.create_instance()
        self.instance.x.add()
        self.instance.x.add()
        self.instance.x.add()
        self.assertEqual(self.instance.x.domain, None)
        self.assertEqual(str(self.instance.x[0].domain), str(NonNegativeReals))
        self.assertEqual(str(self.instance.x[1].domain), str(Reals))
        self.assertEqual(str(self.instance.x[2].domain), str(Integers))
        try:
            self.instance.x.domain = Reals
        except AttributeError:
            pass
        self.assertEqual(self.instance.x.domain, None)

    def test_dim(self):
        """Test dim method"""
        self.model.x = VarList()
        self.instance = self.model.create_instance()
        self.instance.x.add()
        self.assertEqual(self.instance.x.dim(), 1)

    def test_keys(self):
        """Test keys method"""
        self.model.x = VarList()
        self.instance = self.model.create_instance()
        self.instance.x.add()
        self.instance.x.add()
        self.assertEqual(set(self.instance.x.keys()), set([1, 2]))

    def test_len(self):
        """Test len method"""
        self.model.x = VarList()
        self.instance = self.model.create_instance()
        self.instance.x.add()
        self.instance.x.add()
        self.assertEqual(len(self.instance.x), 2)

    def test_value(self):
        """Check the value of the variable"""
        self.model.x = VarList(initialize=3.3)
        self.instance = self.model.create_instance()
        self.instance.x.add()
        self.instance.x.add()
        tmp = value(self.instance.x[1].value)
        self.assertEqual(type(tmp), float)
        self.assertEqual(tmp, 3.3)
        tmp = float(self.instance.x[1].value)
        self.assertEqual(type(tmp), float)
        self.assertEqual(tmp, 3.3)
        tmp = int(self.instance.x[1].value)
        self.assertEqual(type(tmp), int)
        self.assertEqual(tmp, 3)

    def test_0based_add(self):
        m = ConcreteModel()
        m.x = VarList(starting_index=0)
        m.x.add()
        self.assertEqual(list(m.x.keys()), [0])
        m.x.add()
        self.assertEqual(list(m.x.keys()), [0, 1])
        m.x.add()
        self.assertEqual(list(m.x.keys()), [0, 1, 2])

    def test_0based_initialize_with_dict(self):
        """Test initialize option with a dictionary"""
        self.model.x = VarList(initialize={1: 1.3, 2: 2.3}, starting_index=0)
        self.assertRaisesRegex(
            KeyError,
            ".*Index '2' is not valid for indexed component 'x'",
            self.model.create_instance,
        )

    def test_0based_initialize_with_bad_dict(self):
        """Test initialize option with a dictionary of subkeys"""
        self.model.x = VarList(initialize={0: 1.3, 1: 2.3}, starting_index=0)
        self.instance = self.model.create_instance()
        self.assertEqual(self.instance.x[0].value, 1.3)
        self.assertEqual(self.instance.x[1].value, 2.3)
        self.instance.x[0] = 1
        self.instance.x[1] = 2
        self.assertEqual(self.instance.x[0].value, 1)
        self.assertEqual(self.instance.x[1].value, 2)
        self.instance.x.add()
        self.assertEqual(list(self.instance.x.keys()), [0, 1, 2])


class Test2DArrayVar(TestSimpleVar):
    def setUp(self):
        #
        # Create Model
        #
        PyomoModel.setUp(self)
        self.model.A = Set(initialize=[1, 2])

    def test_fixed_attr(self):
        """Test fixed attribute"""
        self.model.x = Var(self.model.A, self.model.A)
        self.model.y = Var(self.model.A, self.model.A)
        self.instance = self.model.create_instance()
        self.instance.x.fixed = True
        self.assertEqual(self.instance.x[1, 2].fixed, False)
        self.instance.y[1, 2].fixed = True
        self.assertEqual(self.instance.y[1, 2].fixed, True)

    def test_value_attr(self):
        """Test value attribute"""
        self.model.x = Var(self.model.A, self.model.A, dense=True)
        self.model.y = Var(self.model.A, self.model.A, dense=True)
        self.instance = self.model.create_instance()
        try:
            self.instance.x = 3.5
            self.fail("Expected ValueError")
        except ValueError:
            pass
        self.instance.y[1, 2] = 3.5
        self.assertEqual(self.instance.y[1, 2].value, 3.5)

    # def test_lb_attr(self):
    # """Test lb attribute"""
    # self.model.x = Var(self.model.A,self.model.A)
    # self.instance = self.model.create_instance()
    # self.instance.x.setlb(-1.0)
    # self.assertEqual(value(self.instance.x[2,1].lb), -1.0)

    # def test_ub_attr(self):
    # """Test ub attribute"""
    # self.model.x = Var(self.model.A,self.model.A)
    # self.instance = self.model.create_instance()
    # self.instance.x.setub(1.0)
    # self.assertEqual(value(self.instance.x[2,1].ub), 1.0)

    def test_initialize_with_function(self):
        """Test initialize option with an initialization rule"""

        def init_rule(model, key1, key2):
            i = key1 + 1
            return key1 == 1 and 1.3 or 2.3

        self.model.x = Var(self.model.A, self.model.A, initialize=init_rule)
        self.instance = self.model.create_instance()
        self.assertEqual(self.instance.x[1, 1].value, 1.3)
        self.assertEqual(self.instance.x[2, 2].value, 2.3)
        self.instance.x[1, 1] = 1
        self.instance.x[2, 2] = 2
        self.assertEqual(self.instance.x[1, 1].value, 1)
        self.assertEqual(self.instance.x[2, 2].value, 2)

    def test_initialize_with_dict(self):
        """Test initialize option with a dictionary"""
        self.model.x = Var(
            self.model.A, self.model.A, initialize={(1, 1): 1.3, (2, 2): 2.3}
        )
        self.instance = self.model.create_instance()
        self.assertEqual(self.instance.x[1, 1].value, 1.3)
        self.assertEqual(self.instance.x[2, 2].value, 2.3)
        self.instance.x[1, 1] = 1
        self.instance.x[2, 2] = 2
        self.assertEqual(self.instance.x[1, 1].value, 1)
        self.assertEqual(self.instance.x[2, 2].value, 2)

    def test_initialize_with_const(self):
        """Test initialize option with a constant"""
        self.model.x = Var(self.model.A, self.model.A, initialize=3)
        self.instance = self.model.create_instance()
        self.assertEqual(self.instance.x[1, 1].value, 3)
        self.assertEqual(self.instance.x[2, 2].value, 3)
        self.instance.x[1, 1] = 1
        self.instance.x[2, 2] = 2
        self.assertEqual(self.instance.x[1, 1].value, 1)
        self.assertEqual(self.instance.x[2, 2].value, 2)

    def test_without_initial_value(self):
        """Test default initialization"""
        self.model.x = Var(self.model.A, self.model.A)
        self.instance = self.model.create_instance()
        self.assertEqual(self.instance.x[1, 1].value, None)
        self.assertEqual(self.instance.x[2, 2].value, None)
        self.instance.x[1, 1] = 5
        self.instance.x[2, 2] = 6
        self.assertEqual(self.instance.x[1, 1].value, 5)
        self.assertEqual(self.instance.x[2, 2].value, 6)

    def test_initialize_option(self):
        """Test initialize option"""
        self.model.x = Var(
            self.model.A, self.model.A, initialize={(1, 1): 1.3, (2, 2): 2.3}
        )
        self.instance = self.model.create_instance()
        self.assertEqual(self.instance.x[1, 1].value, 1.3)
        self.assertEqual(self.instance.x[2, 2].value, 2.3)
        try:
            value(self.instance.x[1, 2])
            self.fail("Expected ValueError")
        except ValueError:
            pass

    def test_bounds_option1(self):
        """Test bounds option"""

        def x_bounds(model, i, j):
            return (-1.0 * (i + j), 1.0 * (i + j))

        self.model.x = Var(self.model.A, self.model.A, bounds=x_bounds)
        self.instance = self.model.create_instance()
        self.assertEqual(value(self.instance.x[1, 1].lb), -2.0)
        self.assertEqual(value(self.instance.x[1, 2].ub), 3.0)

    def test_bounds_option2(self):
        """Test bounds option"""
        self.model.x = Var(self.model.A, self.model.A, bounds=(-1.0, 1.0))
        self.instance = self.model.create_instance()
        self.assertEqual(value(self.instance.x[1, 1].lb), -1.0)
        self.assertEqual(value(self.instance.x[1, 1].ub), 1.0)

    def test_rule_option(self):
        """Test rule option"""

        def x_init(model, i, j):
            return 1.3

        self.model.x = Var(self.model.A, self.model.A, initialize=x_init)
        self.instance = self.model.create_instance()
        self.assertEqual(self.instance.x[1, 2].value, 1.3)

    def test_dim(self):
        """Test dim method"""
        self.model.x = Var(self.model.A, self.model.A)
        self.instance = self.model.create_instance()
        self.assertEqual(self.instance.x.dim(), 2)

    def test_keys(self):
        """Test keys method"""
        self.model.x = Var(self.model.A, self.model.A, dense=True)
        self.instance = self.model.create_instance()
        ans = [(1, 1), (1, 2), (2, 1), (2, 2)]
        self.assertEqual(list(sorted(self.instance.x.keys())), ans)

    def test_len(self):
        """Test len method"""
        self.model.x = Var(self.model.A, self.model.A, dense=True)
        self.instance = self.model.create_instance()
        self.assertEqual(len(self.instance.x), 4)

    def test_value(self):
        """Check the value of the variable"""
        self.model.x = Var(self.model.A, self.model.A, initialize=3.3)
        self.instance = self.model.create_instance()
        tmp = value(self.instance.x[1, 1].value)
        self.assertEqual(type(tmp), float)
        self.assertEqual(tmp, 3.3)
        tmp = float(self.instance.x[1, 1].value)
        self.assertEqual(type(tmp), float)
        self.assertEqual(tmp, 3.3)
        tmp = int(self.instance.x[1, 1].value)
        self.assertEqual(type(tmp), int)
        self.assertEqual(tmp, 3)


class TestVarComplexArray(PyomoModel):
    def test_index1(self):
        self.model.A = Set(initialize=range(0, 4))

        def B_index(model):
            for i in model.A:
                if i % 2 == 0:
                    yield i

        def B_init(model, i, j):
            if j:
                return 2 + i
            return -(2 + i)

        self.model.B = Var(B_index, [True, False], initialize=B_init, dense=True)
        self.instance = self.model.create_instance()
        self.assertEqual(
            set(self.instance.B.keys()),
            set([(0, True), (2, True), (0, False), (2, False)]),
        )
        self.assertEqual(self.instance.B[0, True].value, 2)
        self.assertEqual(self.instance.B[0, False].value, -2)
        self.assertEqual(self.instance.B[2, True].value, 4)
        self.assertEqual(self.instance.B[2, False].value, -4)

    def test_index2(self):
        self.model.A = Set(initialize=range(0, 4))

        def B_index(model):
            for i in model.A:
                if i % 2 == 0:
                    yield i - 1, i

        B_index.dimen = 2

        def B_init(model, k, i, j):
            if j:
                return (2 + i) * k
            return -(2 + i) * k

        self.model.B = Var(B_index, [True, False], initialize=B_init, dense=True)
        self.instance = self.model.create_instance()
        # self.instance.pprint()
        self.assertEqual(
            set(self.instance.B.keys()),
            set([(-1, 0, True), (1, 2, True), (-1, 0, False), (1, 2, False)]),
        )
        self.assertEqual(self.instance.B[-1, 0, True].value, -2)
        self.assertEqual(self.instance.B[-1, 0, False].value, 2)
        self.assertEqual(self.instance.B[1, 2, True].value, 4)
        self.assertEqual(self.instance.B[1, 2, False].value, -4)


class MiscVarTests(unittest.TestCase):
    def test_error1(self):
        a = Var(name="a")
        try:
            a = Var(foo=1)
            self.fail("test_error1")
        except ValueError:
            pass

    def test_getattr1(self):
        """
        Verify the behavior of non-standard suffixes with simple variable
        """
        model = AbstractModel()
        model.a = Var()
        model.suffix = Suffix(datatype=Suffix.INT)
        instance = model.create_instance()
        self.assertEqual(instance.suffix.get(instance.a), None)
        instance.suffix.set_value(instance.a, True)
        self.assertEqual(instance.suffix.get(instance.a), True)

    def test_getattr2(self):
        """
        Verify the behavior of non-standard suffixes with an array of variables
        """
        model = AbstractModel()
        model.X = Set(initialize=[1, 3, 5])
        model.a = Var(model.X)
        model.suffix = Suffix(datatype=Suffix.INT)
        try:
            self.assertEqual(model.a.suffix, None)
            self.fail("Expected AttributeError")
        except AttributeError:
            pass
        instance = model.create_instance()
        self.assertEqual(instance.suffix.get(instance.a[1]), None)
        instance.suffix.set_value(instance.a[1], True)
        self.assertEqual(instance.suffix.get(instance.a[1]), True)

    def test_error2(self):
        try:
            model = AbstractModel()
            model.a = Var(initialize=[1, 2, 3])
            model.b = Var(model.a)
            self.fail("test_error2")
        except TypeError:
            pass

    def test_contains(self):
        model = AbstractModel()
        model.a = Set(initialize=[1, 2, 3])
        model.b = Var(model.a, dense=True)
        instance = model.create_instance()
        self.assertEqual(1 in instance.b, True)

    def test_float_int(self):
        model = AbstractModel()
        model.a = Set(initialize=[1, 2, 3])
        model.b = Var(model.a, initialize=1.1)
        model.c = Var(initialize=2.1)
        model.d = Var()
        instance = model.create_instance()
        self.assertEqual(float(value(instance.b[1])), 1.1)
        self.assertEqual(int(value(instance.b[1])), 1)
        self.assertEqual(float(value(instance.c)), 2.1)
        self.assertEqual(int(value(instance.c)), 2)
        try:
            float(instance.d)
            self.fail("expected TypeError")
        except TypeError:
            pass
        try:
            int(instance.d)
            self.fail("expected TypeError")
        except TypeError:
            pass
        try:
            float(instance.b)
            self.fail("expected TypeError")
        except TypeError:
            pass
        try:
            int(instance.b)
            self.fail("expected TypeError")
        except TypeError:
            pass

    def test_set_get(self):
        model = AbstractModel()
        model.a = Set(initialize=[1, 2, 3])
        model.b = Var(model.a, initialize=1.1, within=PositiveReals)
        model.c = Var(initialize=2.1, within=PositiveReals, bounds=(1, 10))
        with self.assertRaisesRegex(
            ValueError,
            "Cannot set the value for the indexed "
            "component 'b' without specifying an index value",
        ):
            model.b = 2.2

        instance = model.create_instance()
        with self.assertRaisesRegex(
            KeyError, "Cannot treat the scalar component 'c' as an indexed component"
        ):
            instance.c[1] = 2.2

        instance.b[1] = 2.2
        with self.assertRaisesRegex(
            KeyError, "Index '4' is not valid for indexed component 'b'"
        ):
            instance.b[4] = 2.2

        with LoggingIntercept() as LOG:
            instance.b[3] = -2.2
        self.assertEqual(
            LOG.getvalue().strip(),
            "Setting Var 'b[3]' to a value `-2.2` (float) "
            "not in domain PositiveReals.",
        )

        with self.assertRaisesRegex(
            KeyError, "Cannot treat the scalar component 'c' as an indexed component"
        ):
            tmp = instance.c[3]
        with LoggingIntercept() as LOG:
            instance.c = 'a'
        self.assertEqual(
            LOG.getvalue().strip(),
            "Setting Var 'c' to a value `a` (str) not in domain PositiveReals.",
        )
        with LoggingIntercept() as LOG:
            instance.c = -2.2
        self.assertEqual(
            LOG.getvalue().strip(),
            "Setting Var 'c' to a value `-2.2` (float) not in domain PositiveReals.",
        )
        with LoggingIntercept() as LOG:
            instance.c = 11
        self.assertEqual(
            LOG.getvalue().strip(),
            "Setting Var 'c' to a numeric value `11` outside the bounds (1, 10).",
        )

        with LoggingIntercept() as LOG:
            instance.c.set_value('a', skip_validation=True)
        self.assertEqual(LOG.getvalue(), "")
        self.assertEqual(instance.c.value, 'a')
        with LoggingIntercept() as LOG:
            instance.c.set_value(-1, skip_validation=True)
        self.assertEqual(LOG.getvalue(), "")
        self.assertEqual(instance.c.value, -1)

        # try:
        # instance.c.ub = 'a'
        # self.fail("can't set a bad ub for variable c")
        # except ValueError:
        # pass
        # try:
        # instance.c.ub = -1.0
        # self.fail("can't set a bad ub for variable c")
        # except ValueError:
        # pass

        # try:
        # instance.c.fixed = 'a'
        # self.fail("can't fix a variable with a non-boolean")
        # except ValueError:
        # pass

    def test_set_index(self):
        model = ConcreteModel()
        model.s = Set(initialize=[1, 2, 3])
        model.x = Var(model.s, initialize=0, dense=True)

        # test proper instantiation
        self.assertEqual(len(model.x), 3)
        for i in model.s:
            self.assertEqual(value(model.x[i]), 0)

        # test mutability of index set
        model.s.add(4)
        self.assertEqual(len(model.x), 3)
        for i in model.s:
            self.assertEqual(value(model.x[i]), 0)
        self.assertEqual(len(model.x), 4)

    def test_simple_default_domain(self):
        model = ConcreteModel()
        model.x = Var()
        self.assertIs(model.x.domain, Reals)

    def test_simple_nondefault_domain_value(self):
        model = ConcreteModel()
        model.x = Var(domain=Integers)
        self.assertIs(model.x.domain, Integers)

    def test_simple_bad_nondefault_domain_value(self):
        model = ConcreteModel()
        with self.assertRaises(TypeError):
            model.x = Var(domain=25)

    def test_simple_nondefault_domain_rule(self):
        model = ConcreteModel()
        model.x = Var(domain=lambda m: Integers)
        self.assertIs(model.x.domain, Integers)

    def test_simple_bad_nondefault_domain_rule(self):
        model = ConcreteModel()
        with self.assertRaises(TypeError):
            model.x = Var(domain=lambda m: 25)

    def test_indexed_default_domain(self):
        model = ConcreteModel()
        model.s = Set(initialize=[1])
        model.x = Var(model.s)
        self.assertIs(model.x[1].domain, Reals)

    def test_indexed_nondefault_domain_value(self):
        model = ConcreteModel()
        model.s = Set(initialize=[1])
        model.x = Var(model.s, domain=Integers)
        self.assertIs(model.x[1].domain, Integers)

    def test_indexed_bad_nondefault_domain_value(self):
        model = ConcreteModel()
        model.s = Set(initialize=[1])
        with self.assertRaises(TypeError):
            model.x = Var(model.s, domain=25)

    def test_indexed_nondefault_domain_rule(self):
        model = ConcreteModel()
        model.s = Set(initialize=[1])
        model.x = Var(model.s, domain=lambda m, i: Integers)
        self.assertIs(model.x[1].domain, Integers)

    def test_indexed_bad_nondefault_domain_rule(self):
        model = ConcreteModel()
        model.s = Set(initialize=[1])
        with self.assertRaises(TypeError):
            model.x = Var(model.s, domain=lambda m, i: 25)

    def test_list_default_domain(self):
        model = ConcreteModel()
        model.x = VarList()
        model.x.add()
        self.assertIs(model.x[1].domain, Reals)

    def test_list_nondefault_domain_value(self):
        model = ConcreteModel()
        model.x = VarList(domain=Integers)
        model.x.add()
        self.assertIs(model.x[1].domain, Integers)

    def test_list_bad_nondefault_domain_value(self):
        model = ConcreteModel()
        model.x = VarList(domain=25)
        with self.assertRaises(TypeError):
            model.x.add()

    def test_list_nondefault_domain_rule(self):
        model = ConcreteModel()
        model.x = VarList(domain=lambda m, i: Integers)
        model.x.add()
        self.assertIs(model.x[1].domain, Integers)

    def test_list_bad_nondefault_domain_rule(self):
        model = ConcreteModel()
        model.x = VarList(domain=lambda m, i: 25)
        with self.assertRaises(TypeError):
            model.x.add()

    def test_setdata_index(self):
        model = ConcreteModel()
        model.sindex = Set(initialize=[1])
        model.s = Set(model.sindex, initialize=[1, 2, 3])
        model.x = Var(model.s[1], initialize=0, dense=True)

        # test proper instantiation
        self.assertEqual(len(model.x), 3)
        for i in model.s[1]:
            self.assertEqual(value(model.x[i]), 0)

        # test mutability of index set
        newIdx = 4
        self.assertFalse(newIdx in model.s[1])
        self.assertFalse(newIdx in model.x)

        model.s[1].add(newIdx)
        self.assertTrue(newIdx in model.s[1])
        self.assertFalse(newIdx in model.x)

        self.assertEqual(len(model.x), 3)
        for i in model.s[1]:
            self.assertEqual(value(model.x[i]), 0)
        self.assertEqual(len(model.x), 4)

        self.assertTrue(newIdx in model.s[1])
        self.assertTrue(newIdx in model.x)

    def test_setdata_multidimen_index(self):
        model = ConcreteModel()
        model.sindex = Set(initialize=[1])
        model.s = Set(model.sindex, dimen=2, initialize=[(1, 1), (1, 2), (1, 3)])
        model.x = Var(model.s[1], initialize=0, dense=True)

        # test proper instantiation
        self.assertEqual(len(model.x), 3)
        for i in model.s[1]:
            self.assertEqual(value(model.x[i]), 0)

        # test mutability of index set
        newIdx = (1, 4)
        self.assertFalse(newIdx in model.s[1])
        self.assertFalse(newIdx in model.x)

        model.s[1].add(newIdx)
        self.assertTrue(newIdx in model.s[1])
        self.assertFalse(newIdx in model.x)

        self.assertEqual(len(model.x), 3)
        for i in model.s[1]:
            self.assertEqual(value(model.x[i]), 0)
        self.assertEqual(len(model.x), 4)

        self.assertTrue(newIdx in model.s[1])
        self.assertTrue(newIdx in model.x)

    def test_abstract_index(self):
        model = AbstractModel()
        model.A = Set()
        model.B = Set()
        model.C = model.A | model.B
        model.x = Var(model.C)

    @unittest.skipUnless(pint_available, "units test requires pint module")
    def test_set_value_units(self):
        m = ConcreteModel()
        m.x = Var(units=units.g)
        m.x = 5
        self.assertEqual(value(m.x), 5)
        m.x = 6 * units.g
        self.assertEqual(value(m.x), 6)
        m.x = None
        self.assertIsNone(m.x.value, None)
        m.x = 7 * units.kg
        self.assertEqual(value(m.x), 7000)
        with self.assertRaises(UnitsError):
            m.x = 1 * units.s

        out = StringIO()
        m.pprint(ostream=out)
        self.assertEqual(
            out.getvalue().strip(),
            """
1 Var Declarations
    x : Size=1, Index=None, Units=g
        Key  : Lower : Value  : Upper : Fixed : Stale : Domain
        None :  None : 7000.0 :  None : False : False :  Reals

1 Declarations: x
        """.strip(),
        )

    @unittest.skipUnless(pint_available, "units test requires pint module")
    def test_set_bounds_units(self):
        m = ConcreteModel()
        m.x = Var(units=units.g)
        m.p = Param(mutable=True, initialize=1, units=units.kg)
        m.x.setlb(5)
        self.assertEqual(m.x.lb, 5)
        m.x.setlb(6 * units.g)
        self.assertEqual(m.x.lb, 6)
        m.x.setlb(7 * units.kg)
        self.assertEqual(m.x.lb, 7000)
        with self.assertRaises(UnitsError):
            m.x.setlb(1 * units.s)
        m.x.setlb(m.p)
        self.assertEqual(m.x.lb, 1000)
        m.p = 2 * units.kg
        self.assertEqual(m.x.lb, 2000)

        m.x.setub(2)
        self.assertEqual(m.x.ub, 2)
        m.x.setub(3 * units.g)
        self.assertEqual(m.x.ub, 3)
        m.x.setub(4 * units.kg)
        self.assertEqual(m.x.ub, 4000)
        with self.assertRaises(UnitsError):
            m.x.setub(1 * units.s)
        m.x.setub(m.p)
        self.assertEqual(m.x.ub, 2000)
        m.p = 3 * units.kg
        self.assertEqual(m.x.ub, 3000)

    def test_stale(self):
        m = ConcreteModel()
        m.x = Var(initialize=0)
        self.assertFalse(m.x.stale)
        m.y = Var()
        self.assertTrue(m.y.stale)

        StaleFlagManager.mark_all_as_stale(delayed=False)
        self.assertTrue(m.x.stale)
        self.assertTrue(m.y.stale)
        m.x = 1
        self.assertFalse(m.x.stale)
        self.assertTrue(m.y.stale)
        m.y = 2
        self.assertFalse(m.x.stale)
        self.assertFalse(m.y.stale)

        StaleFlagManager.mark_all_as_stale(delayed=True)
        self.assertFalse(m.x.stale)
        self.assertFalse(m.y.stale)
        m.x = 1
        self.assertFalse(m.x.stale)
        self.assertTrue(m.y.stale)
        m.y = 2
        self.assertFalse(m.x.stale)
        self.assertFalse(m.y.stale)

    def test_stale_clone(self):
        m = ConcreteModel()
        m.x = Var(initialize=0)
        self.assertFalse(m.x.stale)
        m.y = Var()
        self.assertTrue(m.y.stale)
        m.z = Var(initialize=0)
        self.assertFalse(m.z.stale)

        i = m.clone()
        self.assertFalse(i.x.stale)
        self.assertTrue(i.y.stale)
        self.assertFalse(i.z.stale)

        StaleFlagManager.mark_all_as_stale(delayed=True)
        m.z = 5
        i = m.clone()
        self.assertTrue(i.x.stale)
        self.assertTrue(i.y.stale)
        self.assertFalse(i.z.stale)

    def test_domain_categories(self):
        """Test domain attribute"""
        x = Var()
        x.construct()
        self.assertEqual(x.is_integer(), False)
        self.assertEqual(x.is_binary(), False)
        self.assertEqual(x.is_continuous(), True)
        self.assertEqual(x.bounds, (None, None))
        x.domain = Integers
        self.assertEqual(x.is_integer(), True)
        self.assertEqual(x.is_binary(), False)
        self.assertEqual(x.is_continuous(), False)
        self.assertEqual(x.bounds, (None, None))
        x.domain = Binary
        self.assertEqual(x.is_integer(), True)
        self.assertEqual(x.is_binary(), True)
        self.assertEqual(x.is_continuous(), False)
        self.assertEqual(x.bounds, (0, 1))
        x.domain = RangeSet(0, 10, 0)
        self.assertEqual(x.is_integer(), False)
        self.assertEqual(x.is_binary(), False)
        self.assertEqual(x.is_continuous(), True)
        self.assertEqual(x.bounds, (0, 10))
        x.domain = RangeSet(0, 10, 1)
        self.assertEqual(x.is_integer(), True)
        self.assertEqual(x.is_binary(), False)
        self.assertEqual(x.is_continuous(), False)
        self.assertEqual(x.bounds, (0, 10))
        x.domain = RangeSet(0.5, 10, 1)
        self.assertEqual(x.is_integer(), False)
        self.assertEqual(x.is_binary(), False)
        self.assertEqual(x.is_continuous(), False)
        self.assertEqual(x.bounds, (0.5, 9.5))
        x.domain = RangeSet(0, 1, 1)
        self.assertEqual(x.is_integer(), True)
        self.assertEqual(x.is_binary(), True)
        self.assertEqual(x.is_continuous(), False)
        self.assertEqual(x.bounds, (0, 1))


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Tests the zero sum propagation module."""

import pyomo.common.unittest as unittest
from pyomo.environ import (
    ConcreteModel,
    Constraint,
    TransformationFactory,
    Var,
    NonNegativeReals,
    NonPositiveReals,
    Binary,
)


class TestZeroSumPropagate(unittest.TestCase):
    """Tests zero sum propagation."""

    def test_non_negative_propagate(self):
        """Test zero sum propagation for non-negative vars."""
        m = ConcreteModel()
        m.v1 = Var(initialize=0)
        m.v2 = Var(initialize=2, domain=NonNegativeReals)
        m.v3 = Var(initialize=3, domain=NonNegativeReals)
        m.v4 = Var(initialize=4, domain=NonNegativeReals)
        m.c1 = Constraint(expr=m.v1 == m.v2 + m.v3 + m.v4)
        m.v1.fix()
        # Because v1 = 0, all the other values must be zero
        TransformationFactory('contrib.propagate_zero_sum').apply_to(m)
        self.assertTrue(m.v1.fixed)
        self.assertTrue(m.v2.fixed)
        self.assertTrue(m.v3.fixed)
        self.assertTrue(m.v4.fixed)
        del m

        m = ConcreteModel()
        m.v1 = Var(initialize=0)
        m.v2 = Var(initialize=2, domain=NonNegativeReals)
        m.v3 = Var(initialize=3, domain=NonNegativeReals)
        m.v4 = Var(initialize=4, domain=NonNegativeReals)
        m.c1 = Constraint(expr=m.v2 + m.v3 + m.v4 == m.v1)
        m.v1.fix()
        TransformationFactory('contrib.propagate_zero_sum').apply_to(m)
        self.assertTrue(m.v1.fixed)
        self.assertTrue(m.v2.fixed)
        self.assertTrue(m.v3.fixed)
        self.assertTrue(m.v4.fixed)
        del m

        m = ConcreteModel()
        m.v1 = Var(initialize=0)
        m.v2 = Var(initialize=2, domain=NonNegativeReals)
        m.v3 = Var(initialize=3, domain=NonNegativeReals)
        m.v4 = Var(initialize=4, domain=NonNegativeReals)
        m.c1 = Constraint(expr=m.v1 >= m.v2 + m.v3 + m.v4)
        m.v1.fix()
        TransformationFactory('contrib.propagate_zero_sum').apply_to(m)
        self.assertTrue(m.v1.fixed)
        self.assertTrue(m.v2.fixed)
        self.assertTrue(m.v3.fixed)
        self.assertTrue(m.v4.fixed)

    def test_not_transform_improperly(self):
        """Tests that invalid constraints are not transformed."""
        m = ConcreteModel()
        m.v1 = Var(initialize=0, domain=Binary)
        m.c1 = Constraint(expr=-1 * m.v1 <= 0)
        TransformationFactory('contrib.propagate_zero_sum').apply_to(m)
        self.assertFalse(m.v1.fixed)

    def test_non_positive_propagate(self):
        """Tests zero sum propagation for non-positive vars."""
        m = ConcreteModel()
        m.v1 = Var(initialize=0)
        m.v2 = Var(initialize=-2, domain=NonPositiveReals)
        m.v3 = Var(initialize=-3, domain=NonPositiveReals)
        m.v4 = Var(initialize=-4, domain=NonPositiveReals)
        m.c1 = Constraint(expr=m.v1 == m.v2 + m.v3 + m.v4)
        m.v1.fix()
        # Because v1 = 0, all the other values must be zero
        TransformationFactory('contrib.propagate_zero_sum').apply_to(m)
        self.assertTrue(m.v1.fixed)
        self.assertTrue(m.v2.fixed)
        self.assertTrue(m.v3.fixed)
        self.assertTrue(m.v4.fixed)
        del m

        m = ConcreteModel()
        m.v1 = Var(initialize=0)
        m.v2 = Var(initialize=-2, domain=NonPositiveReals)
        m.v3 = Var(initialize=-3, domain=NonPositiveReals)
        m.v4 = Var(initialize=-4, domain=NonPositiveReals)
        m.c1 = Constraint(expr=m.v2 + m.v3 + m.v4 == m.v1)
        m.v1.fix()
        TransformationFactory('contrib.propagate_zero_sum').apply_to(m)
        self.assertTrue(m.v1.fixed)
        self.assertTrue(m.v2.fixed)
        self.assertTrue(m.v3.fixed)
        self.assertTrue(m.v4.fixed)
        del m

        m = ConcreteModel()
        m.v1 = Var(initialize=0)
        m.v2 = Var(initialize=-2, domain=NonPositiveReals)
        m.v3 = Var(initialize=-3, domain=NonPositiveReals)
        m.v4 = Var(initialize=-4, domain=NonPositiveReals)
        m.c1 = Constraint(expr=m.v1 <= m.v2 + m.v3 + m.v4)
        m.v1.fix()
        TransformationFactory('contrib.propagate_zero_sum').apply_to(m)
        self.assertTrue(m.v1.fixed)
        self.assertTrue(m.v2.fixed)
        self.assertTrue(m.v3.fixed)
        self.assertTrue(m.v4.fixed)


if __name__ == '__main__':
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.kernel as pmo
from pyomo.core import (
    ConcreteModel,
    Param,
    Var,
    Expression,
    Objective,
    Constraint,
    NonNegativeReals,
)
from pyomo.solvers.tests.models.base import _BaseTestModel, register_model


@register_model
class LP_simple(_BaseTestModel):
    """
    A simple linear program
    """

    description = "LP_simple"
    capabilities = set(['linear'])

    def __init__(self):
        _BaseTestModel.__init__(self)
        self.add_results(self.description + ".json")

    def _generate_model(self):
        self.model = ConcreteModel()
        model = self.model
        model._name = self.description

        model.a1 = Param(initialize=1.0, mutable=True)
        model.a2 = Param([1], initialize=1.0, mutable=True)
        model.a3 = Param(initialize=1.0, mutable=False)
        model.a4 = Param([1], initialize=1.0, mutable=False)
        model.x = Var(within=NonNegativeReals)
        model.y = Var(within=NonNegativeReals)
        model.z1 = Var(bounds=(float('-inf'), float('inf')))
        model.z2 = Var()
        model.dummy_expr1 = Expression(initialize=model.a1 * model.a2[1])
        model.dummy_expr2 = Expression(initialize=model.y / model.a3 * model.a4[1])

        model.inactive_obj = Objective(
            expr=model.x + 3.0 * model.y + 1.0 + model.z1 - model.z2
        )
        model.inactive_obj.deactivate()
        model.p = Param(mutable=True, initialize=0.0)
        model.obj = Objective(expr=model.p + model.inactive_obj)

        model.c1 = Constraint(expr=model.dummy_expr1 <= model.dummy_expr2)
        model.c2 = Constraint(expr=(2.0, model.x / model.a3 - model.y, 10))
        model.c3 = Constraint(expr=(0, model.z1 + 1, 10))
        model.c4 = Constraint(expr=(-10, model.z2 + 1, 0))

    def warmstart_model(self):
        assert self.model is not None
        model = self.model
        model.x.value = None
        model.y.value = 1.0


@register_model
class LP_simple_kernel(LP_simple):
    def _generate_model(self):
        self.model = pmo.block()
        model = self.model
        model._name = self.description

        model.a1 = pmo.parameter(value=1.0)
        model.a2 = pmo.parameter_dict({1: pmo.parameter(value=1.0)})
        model.a3 = pmo.parameter(value=1.0)
        model.a4 = pmo.parameter_dict({1: pmo.parameter(value=1.0)})
        model.x = pmo.variable(domain=NonNegativeReals)
        model.y = pmo.variable(domain=NonNegativeReals)
        model.z1 = pmo.variable()
        model.z2 = pmo.variable()
        model.dummy_expr1 = pmo.expression(model.a1 * model.a2[1])
        model.dummy_expr2 = pmo.expression(model.y / model.a3 * model.a4[1])

        model.inactive_obj = pmo.objective(
            model.x + 3.0 * model.y + 1.0 + model.z1 - model.z2
        )
        model.inactive_obj.deactivate()
        model.p = pmo.parameter(value=0.0)
        model.obj = pmo.objective(model.p + model.inactive_obj)

        model.c1 = pmo.constraint(model.dummy_expr1 <= pmo.noclone(model.dummy_expr2))
        model.c2 = pmo.constraint((2.0, model.x / model.a3 - model.y, 10))
        model.c3 = pmo.constraint((0, model.z1 + 1, 10))
        model.c4 = pmo.constraint((-10, model.z2 + 1, 0))



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.environ as pyo
import pyomo.common.unittest as unittest
from pyomo.common.collections import ComponentSet, ComponentMap
from pyomo.util.subsystems import (
    create_subsystem_block,
    generate_subsystem_blocks,
    TemporarySubsystemManager,
    ParamSweeper,
    identify_external_functions,
    add_local_external_functions,
)
from pyomo.common.gsl import find_GSL


def _make_simple_model():
    m = pyo.ConcreteModel()

    m.v1 = pyo.Var(bounds=(0, None))
    m.v2 = pyo.Var(bounds=(0, None))
    m.v3 = pyo.Var()
    m.v4 = pyo.Var()

    m.con1 = pyo.Constraint(expr=m.v1 * m.v2 * m.v3 == m.v4)
    m.con2 = pyo.Constraint(expr=m.v1 + m.v2**2 == 2 * m.v4)
    m.con3 = pyo.Constraint(expr=m.v1**2 - m.v3 == 3 * m.v4)

    return m


class TestSubsystemBlock(unittest.TestCase):
    def test_square_subsystem(self):
        m = _make_simple_model()

        cons = [m.con2, m.con3]
        vars = [m.v1, m.v2]
        # With m.v3 and m.v4 fixed, m.con2 and m.con3 form a square subsystem
        block = create_subsystem_block(cons, vars)

        self.assertEqual(len(block.vars), 2)
        self.assertEqual(len(block.cons), 2)
        self.assertEqual(len(block.input_vars), 2)
        self.assertEqual(
            len([v for v in block.component_data_objects(pyo.Var) if not v.fixed]), 4
        )

        block.input_vars.fix()
        self.assertEqual(
            len([v for v in block.component_data_objects(pyo.Var) if not v.fixed]), 2
        )

        self.assertIs(block.cons[0], m.con2)
        self.assertIs(block.cons[1], m.con3)
        self.assertIs(block.vars[0], m.v1)
        self.assertIs(block.vars[1], m.v2)
        self.assertIs(block.input_vars[0], m.v4)
        self.assertIs(block.input_vars[1], m.v3)

        # Make sure block is not part of the original model's tree. We
        # don't want to alter the user's model at all.
        self.assertIsNot(block.model(), m)

        # Components on the block are references to components on the
        # original model
        for comp in block.component_objects((pyo.Var, pyo.Constraint)):
            self.assertTrue(comp.is_reference())
            for data in comp.values():
                self.assertIs(data.model(), m)

    def test_subsystem_inputs_only(self):
        m = _make_simple_model()

        cons = [m.con2, m.con3]
        block = create_subsystem_block(cons)

        self.assertEqual(len(block.vars), 0)
        self.assertEqual(len(block.input_vars), 4)
        self.assertEqual(len(block.cons), 2)

        self.assertEqual(
            len([v for v in block.component_data_objects(pyo.Var) if not v.fixed]), 4
        )

        block.input_vars.fix()
        self.assertEqual(
            len([v for v in block.component_data_objects(pyo.Var) if not v.fixed]), 0
        )

        var_set = ComponentSet([m.v1, m.v2, m.v3, m.v4])
        self.assertIs(block.cons[0], m.con2)
        self.assertIs(block.cons[1], m.con3)
        self.assertIn(block.input_vars[0], var_set)
        self.assertIn(block.input_vars[1], var_set)
        self.assertIn(block.input_vars[2], var_set)
        self.assertIn(block.input_vars[3], var_set)

        # Make sure block is not part of the original model's tree. We
        # don't want to alter the user's model at all.
        self.assertIsNot(block.model(), m)

        # Components on the block are references to components on the
        # original model
        for comp in block.component_objects((pyo.Var, pyo.Constraint)):
            self.assertTrue(comp.is_reference())
            for data in comp.values():
                self.assertIs(data.model(), m)

    @unittest.skipUnless(
        pyo.SolverFactory("ipopt").available(), "Ipopt is not available"
    )
    def test_solve_subsystem(self):
        # This is a test of this function's intended use. We extract a
        # subsystem then solve it without altering the rest of the model.
        m = _make_simple_model()
        ipopt = pyo.SolverFactory("ipopt")

        m.v5 = pyo.Var(initialize=1.0)
        m.c4 = pyo.Constraint(expr=m.v5 == 5.0)

        cons = [m.con2, m.con3]
        vars = [m.v1, m.v2]
        block = create_subsystem_block(cons, vars)

        m.v3.fix(1.0)
        m.v4.fix(2.0)

        # Initialize to avoid converging infeasible due to bad pivots
        m.v1.set_value(1.0)
        m.v2.set_value(1.0)
        ipopt.solve(block)

        # Have solved model to expected values
        self.assertAlmostEqual(m.v1.value, pyo.sqrt(7.0), delta=1e-8)
        self.assertAlmostEqual(m.v2.value, pyo.sqrt(4.0 - pyo.sqrt(7.0)), delta=1e-8)

        # Rest of model has not changed
        self.assertEqual(m.v5.value, 1.0)

    def test_generate_subsystems_without_fixed_var(self):
        m = _make_simple_model()
        subs = [([m.con1], [m.v1, m.v4]), ([m.con2, m.con3], [m.v2, m.v3])]
        other_vars = [[m.v2, m.v3], [m.v1, m.v4]]
        for i, (block, inputs) in enumerate(generate_subsystem_blocks(subs)):
            with TemporarySubsystemManager(to_fix=inputs):
                self.assertIs(block.model(), block)
                var_set = ComponentSet(subs[i][1])
                con_set = ComponentSet(subs[i][0])
                input_set = ComponentSet(other_vars[i])

                self.assertEqual(len(var_set), len(block.vars))
                self.assertEqual(len(con_set), len(block.cons))
                self.assertEqual(len(input_set), len(block.input_vars))
                self.assertTrue(all(var in var_set for var in block.vars[:]))
                self.assertTrue(all(con in con_set for con in block.cons[:]))
                self.assertTrue(all(var in input_set for var in inputs))
                self.assertTrue(all(var.fixed for var in inputs))
                self.assertFalse(any(var.fixed for var in block.vars[:]))

        # Test that we have properly unfixed variables
        self.assertFalse(any(var.fixed for var in m.component_data_objects(pyo.Var)))

    def test_generate_subsystems_with_exception(self):
        m = _make_simple_model()
        subs = [([m.con1], [m.v1, m.v4]), ([m.con2, m.con3], [m.v2, m.v3])]
        other_vars = [[m.v2, m.v3], [m.v1, m.v4]]
        with self.assertRaises(RuntimeError):
            for i, (block, inputs) in enumerate(generate_subsystem_blocks(subs)):
                with TemporarySubsystemManager(to_fix=inputs):
                    self.assertTrue(all(var.fixed for var in inputs))
                    self.assertFalse(any(var.fixed for var in block.vars[:]))
                    if i == 1:
                        raise RuntimeError()

        # Test that we have properly unfixed variables
        self.assertFalse(any(var.fixed for var in m.component_data_objects(pyo.Var)))

    def test_generate_subsystems_with_fixed_var(self):
        m = _make_simple_model()
        m.v4.fix()
        subs = [([m.con1], [m.v1]), ([m.con2, m.con3], [m.v2, m.v3])]
        other_vars = [[m.v2, m.v3], [m.v1]]
        for i, (block, inputs) in enumerate(generate_subsystem_blocks(subs)):
            inputs = list(block.input_vars.values())
            with TemporarySubsystemManager(to_fix=inputs):
                self.assertIs(block.model(), block)
                var_set = ComponentSet(subs[i][1])
                con_set = ComponentSet(subs[i][0])
                input_set = ComponentSet(other_vars[i])

                self.assertEqual(len(var_set), len(block.vars))
                self.assertEqual(len(con_set), len(block.cons))
                self.assertEqual(len(input_set), len(inputs))
                self.assertTrue(all(var in var_set for var in block.vars[:]))
                self.assertTrue(all(con in con_set for con in block.cons[:]))
                self.assertTrue(all(var in input_set for var in inputs))
                self.assertTrue(all(var.fixed for var in inputs))
                self.assertFalse(any(var.fixed for var in block.vars[:]))

        # Test that we have properly unfixed variables, except variables
        # that were already fixed.
        self.assertFalse(m.v1.fixed)
        self.assertFalse(m.v2.fixed)
        self.assertFalse(m.v3.fixed)
        self.assertTrue(m.v4.fixed)

    def test_generate_subsystems_include_fixed_var(self):
        m = _make_simple_model()
        m.v4.fix()
        subsystems = [([m.con1], [m.v1]), ([m.con2, m.con3], [m.v2, m.v3])]
        other_vars = [[m.v2, m.v3, m.v4], [m.v1, m.v4]]
        for i, (block, inputs) in enumerate(
            generate_subsystem_blocks(subsystems, include_fixed=True)
        ):
            with TemporarySubsystemManager(to_fix=inputs):
                self.assertIs(block.model(), block)
                var_set = ComponentSet(subsystems[i][1])
                con_set = ComponentSet(subsystems[i][0])
                input_set = ComponentSet(other_vars[i])

                self.assertEqual(len(var_set), len(block.vars))
                self.assertEqual(len(con_set), len(block.cons))
                self.assertEqual(len(input_set), len(block.input_vars))
                self.assertTrue(all(var in var_set for var in block.vars[:]))
                self.assertTrue(all(con in con_set for con in block.cons[:]))
                self.assertTrue(all(var in input_set for var in inputs))
                self.assertTrue(all(var.fixed for var in inputs))
                self.assertFalse(any(var.fixed for var in block.vars[:]))

        self.assertFalse(m.v1.fixed)
        self.assertFalse(m.v2.fixed)
        self.assertFalse(m.v3.fixed)
        self.assertTrue(m.v4.fixed)

    def test_generate_subsystems_dont_fix_inputs(self):
        m = _make_simple_model()
        subs = [([m.con1], [m.v1]), ([m.con2, m.con3], [m.v2, m.v3])]
        other_vars = [[m.v2, m.v3, m.v4], [m.v1, m.v4]]
        for i, (block, inputs) in enumerate(generate_subsystem_blocks(subs)):
            self.assertIs(block.model(), block)
            var_set = ComponentSet(subs[i][1])
            con_set = ComponentSet(subs[i][0])
            input_set = ComponentSet(other_vars[i])

            self.assertEqual(len(var_set), len(block.vars))
            self.assertEqual(len(con_set), len(block.cons))
            self.assertEqual(len(input_set), len(inputs))
            self.assertTrue(all(var in var_set for var in block.vars[:]))
            self.assertTrue(all(con in con_set for con in block.cons[:]))
            self.assertTrue(all(var in input_set for var in inputs))
            self.assertFalse(any(var.fixed for var in inputs))
            self.assertFalse(any(var.fixed for var in block.vars[:]))

        self.assertFalse(m.v1.fixed)
        self.assertFalse(m.v2.fixed)
        self.assertFalse(m.v3.fixed)
        self.assertFalse(m.v4.fixed)

    def test_generate_dont_fix_inputs_with_fixed_var(self):
        m = _make_simple_model()
        m.v4.fix()
        subs = [([m.con1], [m.v1]), ([m.con2, m.con3], [m.v2, m.v3])]
        other_vars = [[m.v2, m.v3], [m.v1]]
        for i, (block, inputs) in enumerate(generate_subsystem_blocks(subs)):
            self.assertIs(block.model(), block)
            var_set = ComponentSet(subs[i][1])
            con_set = ComponentSet(subs[i][0])
            input_set = ComponentSet(other_vars[i])

            self.assertEqual(len(var_set), len(block.vars))
            self.assertEqual(len(con_set), len(block.cons))
            self.assertEqual(len(input_set), len(inputs))
            self.assertTrue(all(var in var_set for var in block.vars[:]))
            self.assertTrue(all(con in con_set for con in block.cons[:]))
            self.assertTrue(all(var in input_set for var in inputs))
            self.assertFalse(m.v1.fixed)
            self.assertFalse(m.v2.fixed)
            self.assertFalse(m.v3.fixed)
            self.assertTrue(m.v4.fixed)

        self.assertFalse(m.v1.fixed)
        self.assertFalse(m.v2.fixed)
        self.assertFalse(m.v3.fixed)
        self.assertTrue(m.v4.fixed)

    def _make_model_with_external_functions(self, named_expressions=False):
        m = pyo.ConcreteModel()
        gsl = find_GSL()
        m.bessel = pyo.ExternalFunction(library=gsl, function="gsl_sf_bessel_J0")
        m.fermi = pyo.ExternalFunction(library=gsl, function="gsl_sf_fermi_dirac_m1")
        m.v1 = pyo.Var(initialize=1.0)
        m.v2 = pyo.Var(initialize=2.0)
        m.v3 = pyo.Var(initialize=3.0)
        if named_expressions:
            m.subexpr = pyo.Expression(pyo.PositiveIntegers)
            m.subexpr[1] = 2 * m.fermi(m.v1)
            m.subexpr[2] = m.bessel(m.v1) - m.bessel(m.v2)
            m.subexpr[3] = m.subexpr[2] + m.v3**2
            subexpr1 = m.subexpr[1]
            subexpr2 = m.subexpr[2]
            subexpr3 = m.subexpr[3]
        else:
            subexpr1 = 2 * m.fermi(m.v1)
            subexpr2 = m.bessel(m.v1) - m.bessel(m.v2)
            subexpr3 = subexpr2 + m.v3**2
        m.con1 = pyo.Constraint(expr=m.v1 == 0.5)
        m.con2 = pyo.Constraint(expr=subexpr1 + m.v2**2 - m.v3 == 1.0)
        m.con3 = pyo.Constraint(expr=subexpr3 == 2.0)
        return m

    @unittest.skipUnless(find_GSL(), "Could not find the AMPL GSL library")
    def test_identify_external_functions(self):
        m = self._make_model_with_external_functions()
        m._con = pyo.Constraint(expr=2 * m.fermi(m.bessel(m.v1**2) + 0.1) == 1.0)

        gsl = find_GSL()

        fcns = list(identify_external_functions(m.con2.expr))
        self.assertEqual(len(fcns), 1)
        self.assertEqual(fcns[0]._fcn._library, gsl)
        self.assertEqual(fcns[0]._fcn._function, "gsl_sf_fermi_dirac_m1")

        fcns = list(identify_external_functions(m.con3.expr))
        fcn_data = set((fcn._fcn._library, fcn._fcn._function) for fcn in fcns)
        self.assertEqual(len(fcns), 2)
        pred_fcn_data = {(gsl, "gsl_sf_bessel_J0")}
        self.assertEqual(fcn_data, pred_fcn_data)

        fcns = list(identify_external_functions(m._con.expr))
        fcn_data = set((fcn._fcn._library, fcn._fcn._function) for fcn in fcns)
        self.assertEqual(len(fcns), 2)
        pred_fcn_data = {(gsl, "gsl_sf_bessel_J0"), (gsl, "gsl_sf_fermi_dirac_m1")}
        self.assertEqual(fcn_data, pred_fcn_data)

    @unittest.skipUnless(find_GSL(), "Could not find the AMPL GSL library")
    def test_local_external_functions_with_named_expressions(self):
        m = self._make_model_with_external_functions(named_expressions=True)
        variables = list(m.component_data_objects(pyo.Var))
        constraints = list(m.component_data_objects(pyo.Constraint, active=True))
        b = create_subsystem_block(constraints, variables)
        self.assertTrue(isinstance(b._gsl_sf_bessel_J0, pyo.ExternalFunction))
        self.assertTrue(isinstance(b._gsl_sf_fermi_dirac_m1, pyo.ExternalFunction))

    def _solve_ef_model_with_ipopt(self):
        m = self._make_model_with_external_functions()
        ipopt = pyo.SolverFactory("ipopt")
        ipopt.solve(m)
        return m

    @unittest.skipUnless(find_GSL(), "Could not find the AMPL GSL library")
    @unittest.skipUnless(
        pyo.SolverFactory("ipopt").available(), "ipopt is not available"
    )
    def test_with_external_function(self):
        m = self._make_model_with_external_functions()
        subsystem = ([m.con2, m.con3], [m.v2, m.v3])

        m.v1.set_value(0.5)
        block = create_subsystem_block(*subsystem)
        ipopt = pyo.SolverFactory("ipopt")
        with TemporarySubsystemManager(to_fix=list(block.input_vars.values())):
            ipopt.solve(block)

        # Correct values obtained by solving with Ipopt directly
        # in another script.
        self.assertEqual(m.v1.value, 0.5)
        self.assertFalse(m.v1.fixed)
        self.assertAlmostEqual(m.v2.value, 1.04816, delta=1e-5)
        self.assertAlmostEqual(m.v3.value, 1.34356, delta=1e-5)

        # Result obtained by solving the full system
        m_full = self._solve_ef_model_with_ipopt()
        self.assertAlmostEqual(m.v1.value, m_full.v1.value)
        self.assertAlmostEqual(m.v2.value, m_full.v2.value)
        self.assertAlmostEqual(m.v3.value, m_full.v3.value)

    @unittest.skipUnless(find_GSL(), "Could not find the AMPL GSL library")
    @unittest.skipUnless(
        pyo.SolverFactory("ipopt").available(), "ipopt is not available"
    )
    def test_with_external_function_in_named_expression(self):
        m = self._make_model_with_external_functions(named_expressions=True)
        subsystem = ([m.con2, m.con3], [m.v2, m.v3])

        m.v1.set_value(0.5)
        block = create_subsystem_block(*subsystem)
        ipopt = pyo.SolverFactory("ipopt")
        with TemporarySubsystemManager(to_fix=list(block.input_vars.values())):
            ipopt.solve(block)

        # Correct values obtained by solving with Ipopt directly
        # in another script.
        self.assertEqual(m.v1.value, 0.5)
        self.assertFalse(m.v1.fixed)
        self.assertAlmostEqual(m.v2.value, 1.04816, delta=1e-5)
        self.assertAlmostEqual(m.v3.value, 1.34356, delta=1e-5)

        # Result obtained by solving the full system
        m_full = self._solve_ef_model_with_ipopt()
        self.assertAlmostEqual(m.v1.value, m_full.v1.value)
        self.assertAlmostEqual(m.v2.value, m_full.v2.value)
        self.assertAlmostEqual(m.v3.value, m_full.v3.value)

    @unittest.skipUnless(find_GSL(), "Could not find the AMPL GSL library")
    def test_external_function_with_potential_name_collision(self):
        m = self._make_model_with_external_functions()
        m.b = pyo.Block()
        m.b._gsl_sf_bessel_J0 = pyo.Var()
        m.b.con = pyo.Constraint(expr=m.b._gsl_sf_bessel_J0 == m.bessel(m.v1))
        add_local_external_functions(m.b)
        self.assertTrue(isinstance(m.b._gsl_sf_bessel_J0, pyo.Var))
        ex_fcns = list(m.b.component_objects(pyo.ExternalFunction))
        self.assertEqual(len(ex_fcns), 1)
        fcn = ex_fcns[0]
        self.assertEqual(fcn._function, "gsl_sf_bessel_J0")


class TestTemporarySubsystemManager(unittest.TestCase):
    def test_context(self):
        m = _make_simple_model()

        to_fix = [m.v4]
        to_deactivate = [m.con1]
        to_reset = [m.v1]

        m.v1.set_value(1.5)

        with TemporarySubsystemManager(to_fix, to_deactivate, to_reset):
            self.assertEqual(m.v1.value, 1.5)
            self.assertTrue(m.v4.fixed)
            self.assertFalse(m.con1.active)

            m.v1.set_value(2.0)
            m.v4.set_value(3.0)

        self.assertEqual(m.v1.value, 1.5)
        self.assertEqual(m.v4.value, 3.0)
        self.assertFalse(m.v4.fixed)
        self.assertTrue(m.con1.active)

    def test_context_some_redundant(self):
        m = _make_simple_model()

        to_fix = [m.v2, m.v4]
        to_deactivate = [m.con1, m.con2]
        to_reset = [m.v1]

        m.v1.set_value(1.5)
        m.v2.fix()
        m.con1.deactivate()

        with TemporarySubsystemManager(to_fix, to_deactivate, to_reset):
            self.assertEqual(m.v1.value, 1.5)
            self.assertTrue(m.v2.fixed)
            self.assertTrue(m.v4.fixed)
            self.assertFalse(m.con1.active)
            self.assertFalse(m.con2.active)

            m.v1.set_value(2.0)
            m.v2.set_value(3.0)

        self.assertEqual(m.v1.value, 1.5)
        self.assertEqual(m.v2.value, 3.0)
        self.assertTrue(m.v2.fixed)
        self.assertFalse(m.v4.fixed)
        self.assertTrue(m.con2.active)
        self.assertFalse(m.con1.active)

    @unittest.skipUnless(
        pyo.SolverFactory("ipopt").available(), "Ipopt is not available"
    )
    def test_fix_then_solve(self):
        # This is a test of the expected use case. We have a (square)
        # subsystem that we can solve easily after fixing and deactivating
        # certain variables and constraints.

        m = _make_simple_model()
        ipopt = pyo.SolverFactory("ipopt")

        # Initialize to avoid converging infeasible due to bad pivots
        m.v1.set_value(1.0)
        m.v2.set_value(1.0)
        m.v3.set_value(1.0)
        m.v4.set_value(2.0)

        with TemporarySubsystemManager(to_fix=[m.v3, m.v4], to_deactivate=[m.con1]):
            # Solve the subsystem with m.v1, m.v2 unfixed and
            # m.con2, m.con3 inactive.
            ipopt.solve(m)

        # Have solved model to expected values
        self.assertAlmostEqual(m.v1.value, pyo.sqrt(7.0), delta=1e-8)
        self.assertAlmostEqual(m.v2.value, pyo.sqrt(4.0 - pyo.sqrt(7.0)), delta=1e-8)

    def test_generate_subsystems_with_exception(self):
        m = _make_simple_model()
        subsystems = [([m.con1], [m.v1, m.v4]), ([m.con2, m.con3], [m.v2, m.v3])]
        other_vars = [[m.v2, m.v3], [m.v1, m.v4]]
        block = create_subsystem_block(*subsystems[0])
        with self.assertRaises(RuntimeError):
            inputs = list(block.input_vars[:])
            with TemporarySubsystemManager(to_fix=inputs):
                self.assertTrue(all(var.fixed for var in inputs))
                self.assertFalse(any(var.fixed for var in block.vars[:]))
                raise RuntimeError()

        # Test that we have properly unfixed variables
        self.assertFalse(any(var.fixed for var in m.component_data_objects(pyo.Var)))

    def test_to_unfix(self):
        m = _make_simple_model()
        m.v1.fix()
        m.v3.fix()
        with TemporarySubsystemManager(to_unfix=[m.v3]):
            self.assertTrue(m.v1.fixed)
            self.assertFalse(m.v2.fixed)
            self.assertFalse(m.v3.fixed)
            self.assertFalse(m.v4.fixed)

        self.assertTrue(m.v1.fixed)
        self.assertFalse(m.v2.fixed)
        self.assertTrue(m.v3.fixed)
        self.assertFalse(m.v4.fixed)


class TestParamSweeper(unittest.TestCase):
    def test_set_values(self):
        m = _make_simple_model()

        n_scenario = 2
        input_values = ComponentMap([(m.v3, [1.3, 2.3]), (m.v4, [1.4, 2.4])])

        to_fix = [m.v3, m.v4]
        to_deactivate = [m.con1]

        with ParamSweeper(
            2, input_values, to_fix=to_fix, to_deactivate=to_deactivate
        ) as sweeper:
            self.assertFalse(m.v1.fixed)
            self.assertFalse(m.v2.fixed)
            self.assertTrue(m.v3.fixed)
            self.assertTrue(m.v4.fixed)
            self.assertFalse(m.con1.active)
            self.assertTrue(m.con2.active)
            self.assertTrue(m.con3.active)
            for i, (inputs, outputs) in enumerate(sweeper):
                self.assertEqual(len(inputs), 2)
                self.assertEqual(len(outputs), 0)
                self.assertIn(m.v3, inputs)
                self.assertIn(m.v4, inputs)
                for var, val in inputs.items():
                    self.assertEqual(var.value, val)
                    self.assertEqual(var.value, input_values[var][i])

        # Values have been reset after exit.
        self.assertIs(m.v3.value, None)
        self.assertIs(m.v4.value, None)

    def test_mutable_parameter(self):
        m = _make_simple_model()
        m.p1 = pyo.Param(mutable=True, initialize=7.0)

        n_scenario = 2
        input_values = ComponentMap(
            [(m.v3, [1.3, 2.3]), (m.v4, [1.4, 2.4]), (m.p1, [1.5, 2.5])]
        )

        to_fix = [m.v3, m.v4]
        to_deactivate = [m.con1]

        with ParamSweeper(
            2, input_values, to_fix=to_fix, to_deactivate=to_deactivate
        ) as sweeper:
            self.assertFalse(m.v1.fixed)
            self.assertFalse(m.v2.fixed)
            self.assertTrue(m.v3.fixed)
            self.assertTrue(m.v4.fixed)
            self.assertFalse(m.con1.active)
            self.assertTrue(m.con2.active)
            self.assertTrue(m.con3.active)
            for i, (inputs, outputs) in enumerate(sweeper):
                self.assertIn(m.v3, inputs)
                self.assertIn(m.v4, inputs)
                self.assertIn(m.p1, inputs)
                self.assertEqual(len(inputs), 3)
                for var, val in inputs.items():
                    self.assertEqual(var.value, val)
                    self.assertEqual(var.value, input_values[var][i])

        # Values have been reset after exit.
        self.assertIs(m.v3.value, None)
        self.assertIs(m.v4.value, None)
        self.assertEqual(m.p1.value, 7.0)

    def test_output_values(self):
        m = _make_simple_model()

        n_scenario = 2
        input_values = ComponentMap([(m.v3, [1.3, 2.3]), (m.v4, [1.4, 2.4])])

        output_values = ComponentMap([(m.v1, [1.1, 2.1]), (m.v2, [1.2, 2.2])])

        to_fix = [m.v3, m.v4]
        to_deactivate = [m.con1]

        with ParamSweeper(
            2, input_values, output_values, to_fix=to_fix, to_deactivate=to_deactivate
        ) as sweeper:
            self.assertFalse(m.v1.fixed)
            self.assertFalse(m.v2.fixed)
            self.assertTrue(m.v3.fixed)
            self.assertTrue(m.v4.fixed)
            self.assertFalse(m.con1.active)
            self.assertTrue(m.con2.active)
            self.assertTrue(m.con3.active)
            for i, (inputs, outputs) in enumerate(sweeper):
                self.assertEqual(len(inputs), 2)
                self.assertEqual(len(outputs), 2)
                self.assertIn(m.v3, inputs)
                self.assertIn(m.v4, inputs)
                self.assertIn(m.v1, outputs)
                self.assertIn(m.v2, outputs)
                for var, val in inputs.items():
                    self.assertEqual(var.value, val)
                    self.assertEqual(var.value, input_values[var][i])

                for var, val in outputs.items():
                    self.assertEqual(val, output_values[var][i])

        # Values have been reset after exit.
        self.assertIs(m.v1.value, None)
        self.assertIs(m.v2.value, None)
        self.assertIs(m.v3.value, None)
        self.assertIs(m.v4.value, None)

    @unittest.skipUnless(
        pyo.SolverFactory("ipopt").available(), "Ipopt is not available"
    )
    def test_with_solve(self):
        m = _make_simple_model()
        ipopt = pyo.SolverFactory("ipopt")

        n_scenario = 2
        input_values = ComponentMap([(m.v3, [1.3, 2.3]), (m.v4, [1.4, 2.4])])

        _v1_val_1 = pyo.sqrt(3 * 1.4 + 1.3)
        _v1_val_2 = pyo.sqrt(3 * 2.4 + 2.3)
        _v2_val_1 = pyo.sqrt(2 * 1.4 - _v1_val_1)
        _v2_val_2 = pyo.sqrt(2 * 2.4 - _v1_val_2)
        output_values = ComponentMap(
            [(m.v1, [_v1_val_1, _v1_val_2]), (m.v2, [_v2_val_1, _v2_val_2])]
        )

        to_fix = [m.v3, m.v4]
        to_deactivate = [m.con1]
        to_reset = [m.v1, m.v2]

        # Initialize values so we don't fail due to bad initialization
        m.v1.set_value(1.0)
        m.v2.set_value(1.0)

        with ParamSweeper(
            n_scenario,
            input_values,
            output_values,
            to_fix=to_fix,
            to_deactivate=to_deactivate,
            to_reset=to_reset,
        ) as sweeper:
            self.assertFalse(m.v1.fixed)
            self.assertFalse(m.v2.fixed)
            self.assertTrue(m.v3.fixed)
            self.assertTrue(m.v4.fixed)
            self.assertFalse(m.con1.active)
            self.assertTrue(m.con2.active)
            self.assertTrue(m.con3.active)
            for i, (inputs, outputs) in enumerate(sweeper):
                ipopt.solve(m)

                for var, val in inputs.items():
                    # These values should not have been altered.
                    # I believe exact equality should be appropriate here.
                    self.assertEqual(var.value, val)
                    self.assertEqual(var.value, input_values[var][i])

                for var, val in outputs.items():
                    self.assertAlmostEqual(var.value, val, delta=1e-8)
                    self.assertAlmostEqual(var.value, output_values[var][i], delta=1e-8)

        # Values have been reset after exit.
        self.assertIs(m.v1.value, 1.0)
        self.assertIs(m.v2.value, 1.0)
        self.assertIs(m.v3.value, None)
        self.assertIs(m.v4.value, None)

    def test_with_exception(self):
        m = _make_simple_model()

        n_scenario = 2
        input_values = ComponentMap([(m.v3, [1.3, 2.3]), (m.v4, [1.4, 2.4])])

        output_values = ComponentMap([(m.v1, [1.1, 2.1]), (m.v2, [1.2, 2.2])])

        to_fix = [m.v3, m.v4]
        to_deactivate = [m.con1]

        with self.assertRaises(RuntimeError):
            with ParamSweeper(
                2,
                input_values,
                output_values,
                to_fix=to_fix,
                to_deactivate=to_deactivate,
            ) as sweeper:
                self.assertFalse(m.v1.fixed)
                self.assertFalse(m.v2.fixed)
                self.assertTrue(m.v3.fixed)
                self.assertTrue(m.v4.fixed)
                self.assertFalse(m.con1.active)
                self.assertTrue(m.con2.active)
                self.assertTrue(m.con3.active)
                for i, (inputs, outputs) in enumerate(sweeper):
                    self.assertEqual(len(inputs), 2)
                    self.assertEqual(len(outputs), 2)
                    self.assertIn(m.v3, inputs)
                    self.assertIn(m.v4, inputs)
                    self.assertIn(m.v1, outputs)
                    self.assertIn(m.v2, outputs)
                    for var, val in inputs.items():
                        self.assertEqual(var.value, val)
                        self.assertEqual(var.value, input_values[var][i])

                    for var, val in outputs.items():
                        self.assertEqual(val, output_values[var][i])

                    if i == 0:
                        raise RuntimeError()

        # Values have been reset after exit.
        self.assertIs(m.v1.value, None)
        self.assertIs(m.v2.value, None)
        self.assertIs(m.v3.value, None)
        self.assertIs(m.v4.value, None)
        self.assertFalse(m.v1.fixed)
        self.assertFalse(m.v2.fixed)
        self.assertFalse(m.v3.fixed)
        self.assertFalse(m.v4.fixed)


if __name__ == '__main__':
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
import numpy as np
from scipy.sparse import coo_matrix
from pyomo.contrib.pynumero.sparse import BlockVector, BlockMatrix
from pyomo.common.dependencies import attempt_import

mpi_block_vector, mpi_block_vector_available = attempt_import(
    'pyomo.contrib.pynumero.sparse.mpi_block_vector'
)


def build_bounds_mask(vector):
    """
    Creates masks for converting from the full vector of bounds that
    may contain -np.inf or np.inf to a vector of bounds that are finite
    only.
    """
    return build_compression_mask_for_finite_values(vector)


def build_compression_matrix(compression_mask):
    """
    Return a sparse matrix CM of ones such that
    compressed_vector = CM*full_vector based on the
    compression mask

    Parameters
    ----------
    compression_mask: np.ndarray or pyomo.contrib.pynumero.sparse.block_vector.BlockVector

    Returns
    -------
    cm: coo_matrix or BlockMatrix
       The compression matrix
    """
    if isinstance(compression_mask, BlockVector):
        n = compression_mask.nblocks
        res = BlockMatrix(nbrows=n, nbcols=n)
        for ndx, block in enumerate(compression_mask):
            sub_matrix = build_compression_matrix(block)
            res.set_block(ndx, ndx, sub_matrix)
        return res
    elif type(compression_mask) is np.ndarray:
        cols = compression_mask.nonzero()[0]
        nnz = len(cols)
        rows = np.arange(nnz, dtype=np.int64)
        data = np.ones(nnz)
        return coo_matrix((data, (rows, cols)), shape=(nnz, len(compression_mask)))
    elif isinstance(compression_mask, mpi_block_vector.MPIBlockVector):
        from pyomo.contrib.pynumero.sparse.mpi_block_matrix import MPIBlockMatrix

        n = compression_mask.nblocks
        rank_ownership = np.ones((n, n), dtype=np.int64) * -1
        for i in range(n):
            rank_ownership[i, i] = compression_mask.rank_ownership[i]
        res = MPIBlockMatrix(
            nbrows=n,
            nbcols=n,
            rank_ownership=rank_ownership,
            mpi_comm=compression_mask.mpi_comm,
            assert_correct_owners=False,
        )
        for ndx in compression_mask.owned_blocks:
            block = compression_mask.get_block(ndx)
            sub_matrix = build_compression_matrix(block)
            res.set_block(ndx, ndx, sub_matrix)
        return res


def build_compression_mask_for_finite_values(vector):
    """
    Creates masks for converting from the full vector of
    values to the vector that contains only the finite values. This is
    typically used to convert a vector of bounds (that may contain np.inf
    and -np.inf) to only the bounds that are finite.
    """
    full_finite_mask = np.isfinite(vector)
    return full_finite_mask


# TODO: Is this needed anywhere?
# def build_expansion_map_for_finite_values(vector):
#    """
#    Creates a map from the compressed vector to the full
#    vector based on the locations of finite values only. This is
#    typically used to map a vector of bounds (that is compressed to only
#    contain the finite values) to a full vector (that may contain np.inf
#    and -np.inf).
#    """
#    full_finite_mask = np.isfinite(vector)
#    finite_full_map = full_finite_mask.nonzero()[0]
#    return finite_full_map


def full_to_compressed(full_array, compression_mask, out=None):
    if out is not None:
        np.compress(compression_mask, full_array, out=out)
        return out
    else:
        return np.compress(compression_mask, full_array)


def compressed_to_full(compressed_array, compression_mask, out=None, default=None):
    if out is None:
        ret = np.empty(len(compression_mask))
        ret.fill(np.nan)
    else:
        ret = out

    ret[compression_mask] = compressed_array
    if default is not None:
        ret[~compression_mask] = default

    return ret


def make_lower_triangular_full(lower_triangular_matrix):
    '''
    This function takes a symmetric matrix that only has entries in the
    lower triangle and makes is a full matrix by duplicating the entries
    '''
    mask = lower_triangular_matrix.row != lower_triangular_matrix.col

    row = np.concatenate(
        (lower_triangular_matrix.row, lower_triangular_matrix.col[mask])
    )
    col = np.concatenate(
        (lower_triangular_matrix.col, lower_triangular_matrix.row[mask])
    )
    data = np.concatenate(
        (lower_triangular_matrix.data, lower_triangular_matrix.data[mask])
    )

    return coo_matrix((data, (row, col)), shape=lower_triangular_matrix.shape)


class CondensedSparseSummation:
    def __init__(self, list_of_matrices):
        """
        This class is used to perform a summation of sparse matrices
        while retaining the correct and consistent nonzero structure.
        Create the class with the list of matrices you want to sum,
        and the condensed_summation method remains valid as long as
        the structure of the individual matrices is consistent
        """
        self._nz_tuples = None
        self._maps = None
        self._build_maps(list_of_matrices)

    def _build_maps(self, list_of_matrices):
        """
        This method creates the maps that are used in condensed_sum.
        These maps remain valid as long as the nonzero structure of
        the individual matrices does not change
        """
        # get the list of all unique nonzeros across the matrices
        nz_tuples = set()
        for m in list_of_matrices:
            nz_tuples.update(zip(m.row, m.col))
        nz_tuples = sorted(nz_tuples)
        self._nz_tuples = nz_tuples
        self._row, self._col = list(zip(*nz_tuples))
        row_col_to_nz_map = {t: i for i, t in enumerate(nz_tuples)}

        self._shape = None
        self._maps = list()
        for m in list_of_matrices:
            nnz = len(m.data)
            map_row = np.zeros(nnz)
            map_col = np.zeros(nnz)
            for i in range(nnz):
                map_col[i] = i
                map_row[i] = row_col_to_nz_map[(m.row[i], m.col[i])]
            mp = coo_matrix(
                (np.ones(nnz), (map_row, map_col)), shape=(len(row_col_to_nz_map), nnz)
            )
            self._maps.append(mp)
            if self._shape is None:
                self._shape = m.shape
            else:
                assert self._shape == m.shape

    def sum(self, list_of_matrices):
        data = np.zeros(len(self._row))
        assert len(self._maps) == len(list_of_matrices)
        for i, mp in enumerate(self._maps):
            data += mp.dot(list_of_matrices[i].data)
        ret = coo_matrix(
            (data, (np.copy(self._row), np.copy(self._col))), shape=self._shape
        )
        return ret



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.config import ConfigBlock, ConfigList, ConfigValue


class Default_Config:
    def config_block(self, init=False):
        config, blocks = minlp_config_block(init=init)
        return config, blocks


def minlp_config_block(init=False):
    config = ConfigBlock(
        "Configuration for a canonical model construction and optimization sequence"
    )
    blocks = {}

    #
    # Data
    #
    data = config.declare('data', ConfigBlock())
    blocks['data'] = data

    data.declare(
        'files',
        ConfigList(
            [], ConfigValue(None, str, 'Filename', None), 'Model data files', None
        ),
    )
    data.declare(
        'namespaces',
        ConfigList(
            [],
            ConfigValue(None, str, 'Namespace', None),
            'A namespace that is used to select data in Pyomo data files.',
            None,
        ),
    ).declare_as_argument('--namespace', dest='namespaces', action='append')

    #
    # Model
    #
    model = config.declare('model', ConfigBlock())
    blocks['model'] = model

    model.declare(
        'filename',
        ConfigValue(None, str, 'The Python module that specifies the model', None),
    )
    model.declare(
        'object name',
        ConfigValue(
            None,
            str,
            'The name of the model object that is created in the '
            'specified Pyomo module',
            None,
        ),
    ).declare_as_argument('--model-name', dest='model_name')
    model.declare('type', ConfigValue(None, str, 'The problem type', None))
    model.declare(
        'options',
        ConfigBlock(implicit=True, description='Options used to construct the model'),
    )
    model.declare(
        'linearize expressions',
        ConfigValue(
            False,
            bool,
            'An option intended for use on linear or mixed-integer models '
            'in which expression trees in a model (constraints or objectives) '
            'are compacted into a more memory-efficient and concise form.',
            None,
        ),
    )
    model.declare(
        'save file',
        ConfigValue(
            None,
            str,
            'The filename to which the model is saved. The suffix of this '
            'filename specifies the file format.',
            None,
        ),
    )
    model.declare(
        'save format',
        ConfigValue(
            None,
            str,
            "The format that the model is saved. When specified, this "
            "overrides the format implied by the 'save file' option.",
            None,
        ),
    )
    model.declare(
        'symbolic solver labels',
        ConfigValue(
            False,
            bool,
            'When interfacing with the solver, use symbol names derived '
            'from the model. For example, "my_special_variable[1_2_3]" '
            'instead of "v1". Useful for debugging. When using the ASL '
            'interface (--solver-io=nl), generates corresponding .row '
            '(constraints) and .col (variables) files. The ordering in '
            'these files provides a mapping from ASL index to symbolic '
            'model names.',
            None,
        ),
    ).declare_as_argument(dest='symbolic_solver_labels')
    model.declare(
        'file determinism',
        ConfigValue(
            None,
            int,
            'When interfacing with a solver using file based I/O, set '
            'the effort level for ensuring the file creation process is '
            'deterministic. See the individual solver interfaces for '
            'valid values and default level of file determinism.',
            None,
        ),
    ).declare_as_argument(dest='file_determinism')

    #
    # Transform
    #
    transform = ConfigBlock()
    blocks['transform'] = transform

    transform.declare(
        'name', ConfigValue(None, str, 'Name of the model transformation', None)
    )
    transform.declare(
        'options', ConfigBlock(implicit=True, description='Transformation options')
    )
    #
    transform_list = config.declare(
        'transform',
        ConfigList(
            [],
            ConfigValue(None, str, 'Transformation', None),
            'List of model transformations',
            None,
        ),
    ).declare_as_argument(dest='transformations', action='append')
    if init:
        transform_list.append()

    #
    # Preprocess
    #
    config.declare(
        'preprocess',
        ConfigList(
            [],
            ConfigValue(None, str, 'Module', None),
            'Specify a Python module that gets immediately executed '
            '(before the optimization model is setup).',
            None,
        ),
    ).declare_as_argument(dest='preprocess')

    #
    # Runtime
    #
    runtime = config.declare('runtime', ConfigBlock())
    blocks['runtime'] = runtime

    runtime.declare(
        'logging',
        ConfigValue(
            None, str, 'Logging level:  quiet, warning, info, verbose, debug', None
        ),
    ).declare_as_argument(dest="logging", metavar="LEVEL")
    runtime.declare(
        'logfile',
        ConfigValue(None, str, 'Redirect output to the specified file.', None),
    ).declare_as_argument(dest="output", metavar="FILE")
    runtime.declare(
        'catch errors',
        ConfigValue(
            False,
            bool,
            'Trigger failures for exceptions to print the program stack.',
            None,
        ),
    ).declare_as_argument('-c', '--catch-errors', dest="catch")
    runtime.declare(
        'disable gc', ConfigValue(False, bool, 'Disable the garbage collector.', None)
    ).declare_as_argument('--disable-gc', dest='disable_gc')
    runtime.declare(
        'interactive',
        ConfigValue(
            False,
            bool,
            'After executing Pyomo, launch an interactive Python shell. '
            'If IPython is installed, this shell is an IPython shell.',
            None,
        ),
    )
    runtime.declare(
        'keep files', ConfigValue(False, bool, 'Keep temporary files', None)
    ).declare_as_argument('-k', '--keepfiles', dest='keepfiles')
    runtime.declare(
        'paths',
        ConfigList(
            [],
            ConfigValue(None, str, 'Path', None),
            'Give a path that is used to find the Pyomo python files.',
            None,
        ),
    ).declare_as_argument('--path', dest='path')
    runtime.declare(
        'profile count',
        ConfigValue(
            0,
            int,
            'Enable profiling of Python code. The value of this option '
            'is the number of functions that are summarized.',
            None,
        ),
    ).declare_as_argument(dest='profile_count', metavar='COUNT')
    runtime.declare(
        'profile memory',
        ConfigValue(
            0,
            int,
            "Report memory usage statistics for the generated instance "
            "and any associated processing steps. A value of 0 indicates "
            "disabled. A value of 1 forces the print of the total memory "
            "after major stages of the pyomo script. A value of 2 forces "
            "summary memory statistics after major stages of the pyomo "
            "script. A value of 3 forces detailed memory statistics "
            "during instance creation and various steps of preprocessing. "
            "Values equal to 4 and higher currently provide no additional "
            "information. Higher values automatically enable all "
            "functionality associated with lower values, e.g., 3 turns "
            "on detailed and summary statistics.",
            None,
        ),
    )
    runtime.declare(
        'report timing',
        ConfigValue(
            False,
            bool,
            'Report various timing statistics during model construction.',
            None,
        ),
    ).declare_as_argument(dest='report_timing')
    runtime.declare(
        'tempdir',
        ConfigValue(
            None,
            str,
            'Specify the directory where temporary files are generated.',
            None,
        ),
    ).declare_as_argument(dest='tempdir')

    return config, blocks


def default_config_block(solver, init=False):
    config, blocks = Default_Config().config_block(init)

    #
    # Solver
    #
    solver = ConfigBlock()
    solver.declare('solver name', ConfigValue('glpk', str, 'Solver name', None))
    solver.declare(
        'solver executable',
        ConfigValue(
            default=None,
            domain=str,
            description="The solver executable used by the solver interface.",
            doc=(
                "The solver executable used by the solver interface. "
                "This option is only valid for those solver interfaces that "
                "interact with a local executable through the shell. If unset, "
                "the solver interface will attempt to find an executable within "
                "the search path of the shell's environment that matches a name "
                "commonly associated with the solver interface."
            ),
        ),
    )
    solver.declare(
        'io format',
        ConfigValue(
            None,
            str,
            'The type of IO used to execute the solver. Different solvers '
            'support different types of IO, but the following are common '
            'options: lp - generate LP files, nl - generate NL files, '
            'python - direct Python interface, os - generate OSiL XML files.',
            None,
        ),
    )
    solver.declare(
        'manager',
        ConfigValue(
            'serial',
            str,
            'The technique that is used to manage solver executions.',
            None,
        ),
    )
    solver.declare(
        'options',
        ConfigBlock(
            implicit=True,
            implicit_domain=ConfigValue(None, str, 'Solver option', None),
            description="Options passed into the solver",
        ),
    )
    solver.declare(
        'options string',
        ConfigValue(None, str, 'String describing solver options', None),
    )
    solver.declare(
        'suffixes',
        ConfigList(
            [],
            ConfigValue(None, str, 'Suffix', None),
            'Solution suffixes that will be extracted by the solver '
            '(e.g., rc, dual, or slack). The use of this option is not '
            'required when a suffix has been declared on the model '
            'using Pyomo\'s Suffix component.',
            None,
        ),
    )
    blocks['solver'] = solver
    #
    solver_list = config.declare(
        'solvers',
        ConfigList(
            [],
            solver,  # ConfigValue(None, str, 'Solver', None),
            'List of solvers.  The first solver in this list is the main solver.',
            None,
        ),
    )
    #
    # Make sure that there is one solver in the list.
    #
    # This will be the solver into which we dump command line options.
    # Note that we CANNOT declare the argparse options on the base block
    # definition above, as we use that definition as the DOMAIN TYPE for
    # the list of solvers.  As that information is NOT copied to
    # derivative blocks, the initial solver entry we are creating would
    # be missing all argparse information. Plus, if we were to have more
    # than one solver defined, we wouldn't want command line options
    # going to both.
    solver_list.append()
    # solver_list[0].get('solver name').\
    #    declare_as_argument('--solver', dest='solver')
    solver_list[0].get('solver executable').declare_as_argument(
        '--solver-executable', dest="solver_executable", metavar="FILE"
    )
    solver_list[0].get('io format').declare_as_argument(
        '--solver-io', dest='io_format', metavar="FORMAT"
    )
    solver_list[0].get('manager').declare_as_argument(
        '--solver-manager', dest="smanager_type", metavar="TYPE"
    )
    solver_list[0].get('options string').declare_as_argument(
        '--solver-options', dest='options_string', metavar="STRING"
    )
    solver_list[0].get('suffixes').declare_as_argument(
        '--solver-suffix', dest="solver_suffixes"
    )

    #
    # Postprocess
    #
    config.declare(
        'postprocess',
        ConfigList(
            [],
            ConfigValue(None, str, 'Module', None),
            'Specify a Python module that gets executed after optimization.',
            None,
        ),
    ).declare_as_argument(dest='postprocess')

    #
    # Postsolve
    #
    postsolve = config.declare('postsolve', ConfigBlock())
    blocks['postsolve'] = postsolve

    postsolve.declare(
        'print logfile',
        ConfigValue(
            False, bool, 'Print the solver logfile after performing optimization.', None
        ),
    ).declare_as_argument('-l', '--log', dest="log")
    postsolve.declare(
        'save results',
        ConfigValue(
            None, str, 'Specify the filename to which the results are saved.', None
        ),
    ).declare_as_argument('--save-results', dest="save_results", metavar="FILE")
    postsolve.declare(
        'show results',
        ConfigValue(False, bool, 'Print the results object after optimization.', None),
    ).declare_as_argument(dest="show_results")
    postsolve.declare(
        'results format',
        ConfigValue(None, str, 'Specify the results format:  json or yaml.', None),
    ).declare_as_argument(
        '--results-format', dest="results_format", metavar="FORMAT"
    ).declare_as_argument(
        '--json',
        dest="results_format",
        action="store_const",
        const="json",
        help="Store results in JSON format",
    )
    postsolve.declare(
        'summary',
        ConfigValue(
            False,
            bool,
            'Summarize the final solution after performing optimization.',
            None,
        ),
    ).declare_as_argument(dest="summary")

    #
    # Runtime
    #
    runtime = blocks['runtime']
    runtime.declare(
        'only instance',
        ConfigValue(False, bool, "Generate a model instance, and then exit", None),
    ).declare_as_argument('--instance-only', dest='only_instance')
    runtime.declare(
        'stream output',
        ConfigValue(
            False,
            bool,
            "Stream the solver output to provide information about the "
            "solver's progress.",
            None,
        ),
    ).declare_as_argument('--stream-output', '--stream-solver', dest="tee")
    #
    return config, blocks



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import math
import pyomo.common.unittest as unittest
from pyomo.contrib.solver.common.factory import SolverFactory
from pyomo.contrib.solver.common.results import TerminationCondition, SolutionStatus
from pyomo.contrib.solver.solvers.gurobi.gurobi_direct_minlp import GurobiDirectMINLP

from pyomo.core.base.constraint import Constraint
from pyomo.environ import (
    ConcreteModel,
    Var,
    VarList,
    Constraint,
    ConstraintList,
    value,
    Binary,
    NonNegativeReals,
    Objective,
    maximize,
    minimize,
    cos,
    sin,
    tan,
    log,
    log10,
    exp,
    sqrt,
)

gurobi_direct = SolverFactory('gurobi_direct_minlp')


@unittest.skipUnless(gurobi_direct.available(), "needs Gurobi Direct MINLP interface")
@unittest.pytest.mark.solver("gurobi_direct_minlp")
class TestGurobiMINLP(unittest.TestCase):
    def test_gurobi_minlp_sincosexp(self):
        m = ConcreteModel(name="test")
        m.x = Var(bounds=(-1, 4))
        m.o = Objective(expr=sin(m.x) + cos(2 * m.x) + 1)
        m.c = Constraint(expr=0.25 * exp(m.x) - m.x <= 0)
        gurobi_direct.solve(m)
        self.assertAlmostEqual(1, value(m.o), delta=1e-3)
        self.assertAlmostEqual(1.571, value(m.x), delta=1e-3)

    def test_gurobi_minlp_tan(self):
        m = ConcreteModel(name="test")
        m.x = Var(bounds=(0, math.pi / 2))
        m.o = Objective(expr=tan(m.x) / (m.x**2))
        gurobi_direct.solve(m)
        self.assertAlmostEqual(0.948, value(m.x), delta=1e-3)
        self.assertAlmostEqual(1.549, value(m.o), delta=1e-3)

    def test_gurobi_minlp_sqrt(self):
        m = ConcreteModel(name="test")
        m.x = Var(bounds=(0, 2))
        m.o = Objective(expr=sqrt(m.x) - (m.x**2) / 3, sense=maximize)
        gurobi_direct.solve(m)
        self.assertAlmostEqual(0.825, value(m.x), delta=1e-3)
        self.assertAlmostEqual(0.681, value(m.o), delta=1e-3)

    def test_gurobi_minlp_log(self):
        m = ConcreteModel(name="test")
        m.x = Var(bounds=(1, 2))
        m.o = Objective(expr=(m.x * m.x) / log(m.x))
        gurobi_direct.solve(m)
        self.assertAlmostEqual(sqrt(math.e), value(m.x), delta=1e-3)
        self.assertAlmostEqual(2 * math.e, value(m.o), delta=1e-3)

    def test_gurobi_minlp_log10(self):
        m = ConcreteModel(name="test")
        m.x = Var(bounds=(1, 2))
        m.o = Objective(expr=(m.x * m.x) / log10(m.x))
        gurobi_direct.solve(m)
        self.assertAlmostEqual(1.649, value(m.x), delta=1e-3)
        self.assertAlmostEqual(12.518, value(m.o), delta=1e-3)

    def test_gurobi_minlp_sigmoid(self):
        m = ConcreteModel(name="test")
        m.x = Var(bounds=(0, 4))
        m.y = Var(bounds=(0, 4))
        m.o = Objective(expr=m.y - m.x)
        m.c = Constraint(expr=1 / (1 + exp(-m.x)) <= m.y)
        gurobi_direct.solve(m)
        self.assertAlmostEqual(value(m.o), -3.017, delta=1e-3)

    def _build_divpwr_model(self, divide: bool, min: bool):
        model = ConcreteModel(name="test")
        model.x1 = Var(domain=NonNegativeReals, bounds=(0.5, 0.6))
        model.x2 = Var(domain=NonNegativeReals, bounds=(0.1, 0.2))
        model.y = Var(domain=Binary, initialize=1)

        y_mult = 1.3
        if divide:
            obj = (1 - model.y) / model.x1 + model.y * y_mult / model.x2
        else:
            obj = (1 - model.y) * (model.x1**-1) + model.y * y_mult * (model.x2**-1)

        if min:
            model.OBJ = Objective(expr=-1 * obj, sense=minimize)
        else:
            model.OBJ = Objective(expr=obj, sense=maximize)

        return model

    def test_gurobi_minlp_divpwr(self):
        params = [
            {"min": False, "divide": False, "obj": 13},
            {"min": False, "divide": True, "obj": 13},
            {"min": True, "divide": False, "obj": -13},
            {"min": True, "divide": True, "obj": -13},
        ]
        for p in params:
            model = self._build_divpwr_model(p['divide'], p['min'])
            gurobi_direct.solve(model)
            self.assertEqual(p["obj"], value(model.OBJ))
            self.assertEqual(1, model.y.value)

    def test_gurobi_minlp_acopf(self):
        # Based on https://docs.gurobi.com/projects/examples/en/current/examples/python/acopf_4buses.html

        # Number of Buses (Nodes)
        N = 4

        # Conductance/susceptance components
        G = [
            [1.7647, -0.5882, 0.0, -1.1765],
            [-0.5882, 1.5611, -0.3846, -0.5882],
            [0.0, -0.3846, 1.5611, -1.1765],
            [-1.1765, -0.5882, -1.1765, 2.9412],
        ]
        B = [
            [-7.0588, 2.3529, 0.0, 4.7059],
            [2.3529, -6.629, 1.9231, 2.3529],
            [0.0, 1.9231, -6.629, 4.7059],
            [4.7059, 2.3529, 4.7059, -11.7647],
        ]

        # Assign bounds where fixings are needed
        v_lb = [1.0, 0.0, 1.0, 0.0]
        v_ub = [1.0, 1.5, 1.0, 1.5]
        P_lb = [-3.0, -0.3, 0.3, -0.2]
        P_ub = [3.0, -0.3, 0.3, -0.2]
        Q_lb = [-3.0, -0.2, -3.0, -0.15]
        Q_ub = [3.0, -0.2, 3.0, -0.15]
        theta_lb = [0.0, -math.pi / 2, -math.pi / 2, -math.pi / 2]
        theta_ub = [0.0, math.pi / 2, math.pi / 2, math.pi / 2]

        exp_v = [1.0, 0.949, 1.0, 0.973]
        exp_theta = [0.0, -2.176, 1.046, -0.768]
        exp_P = [0.2083, -0.3, 0.3, -0.2]
        exp_Q = [0.212, -0.2, 0.173, -0.15]

        m = ConcreteModel(name="acopf")

        m.P = VarList()
        m.Q = VarList()
        m.v = VarList()
        m.theta = VarList()

        for i in range(N):
            p = m.P.add()
            p.lb = P_lb[i]
            p.ub = P_ub[i]

            q = m.Q.add()
            q.lb = Q_lb[i]
            q.ub = Q_ub[i]

            v = m.v.add()
            v.lb = v_lb[i]
            v.ub = v_ub[i]

            theta = m.theta.add()
            theta.lb = theta_lb[i]
            theta.ub = theta_ub[i]

        m.obj = Objective(expr=m.Q[1] + m.Q[3])

        m.define_P = ConstraintList()
        m.define_Q = ConstraintList()
        for i in range(N):
            m.define_P.add(
                m.P[i + 1]
                == m.v[i + 1]
                * sum(
                    m.v[j + 1]
                    * (
                        G[i][j] * cos(m.theta[i + 1] - m.theta[j + 1])
                        + B[i][j] * sin(m.theta[i + 1] - m.theta[j + 1])
                    )
                    for j in range(N)
                )
            )
            m.define_Q.add(
                m.Q[i + 1]
                == m.v[i + 1]
                * sum(
                    m.v[j + 1]
                    * (
                        G[i][j] * sin(m.theta[i + 1] - m.theta[j + 1])
                        - B[i][j] * cos(m.theta[i + 1] - m.theta[j + 1])
                    )
                    for j in range(N)
                )
            )

        results = gurobi_direct.solve(m, tee=True)
        self.assertEqual(
            results.termination_condition,
            TerminationCondition.convergenceCriteriaSatisfied,
        )
        self.assertEqual(results.solution_status, SolutionStatus.optimal)
        for i in range(N):
            self.assertAlmostEqual(
                exp_P[i], m.P[i + 1].value, delta=1e-3, msg=f'P[{i}]'
            )
            self.assertAlmostEqual(
                exp_Q[i], m.Q[i + 1].value, delta=1e-3, msg=f'Q[{i}]'
            )
            self.assertAlmostEqual(
                exp_v[i], m.v[i + 1].value, delta=1e-3, msg=f'v[{i}]'
            )
            self.assertAlmostEqual(
                exp_theta[i],
                m.theta[i + 1].value * 180 / math.pi,
                delta=1e-3,
                msg=f'theta[{i}]',
            )
